import numpy as np
import openmdao.api as om
from openmdao.utils import cs_safe as cs

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.aerodynamics.gasp_based.common import (AeroForces,
                                                              CLFromLift,
                                                              TanhRampComp)
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic, Mission

#
# data from INTERFERENCE - polynomial coefficients
#
ckl = np.array([0.13077, 1.9791, 3.3325, -10.095, 4.7229])
ckv = np.array([0.0194, -0.14817, 1.3515])
ckdt = np.array([0.73543, 0.028571])

#
# data from EAERO
#
sig1 = np.array(
    [
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.2, 0.1833, 0.1621, 0.1429, 0.1256, 0.1101, 0.0966],
        [0.4, 0.3600, 0.3186, 0.2801, 0.2454, 0.2147, 0.1879],
        [0.6, 0.5319, 0.4654, 0.4053, 0.3526, 0.3070, 0.2681],
        [0.8, 0.6896, 0.5900, 0.5063, 0.4368, 0.3791, 0.3309],
        [1.0, 0.7857, 0.6575, 0.5613, 0.4850, 0.4228, 0.3712],
    ]
)
sig2 = np.array(
    [
        [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
        [0.04, 0.7971, 0.9314, 0.9722, 0.9874, 0.9939, 0.9969],
        [0.16, 0.7040, 0.8681, 0.9373, 0.9688, 0.9839, 0.9914],
        [0.36, 0.7476, 0.8767, 0.9363, 0.9659, 0.9812, 0.9893],
        [0.64, 0.8709, 0.9338, 0.9625, 0.9778, 0.9865, 0.9917],
        [1.0, 0.9852, 0.9852, 0.9880, 0.9910, 0.9935, 0.9954],
    ]
)
xbbar = np.linspace(0, 1.0, sig1.shape[0])
xhbar = np.linspace(0, 0.3, sig1.shape[1])

#
# data from DRAG
#
# flap deflection angles, deg
adelfd = np.array(
    [
        0.0,
        5.0,
        10.0,
        15.0,
        20.0,
        25.0,
        30.0,
        35.0,
        38.0,
        40.0,
        42.0,
        44.0,
        50.0,
        55.0,
        60.0,
    ]
)
# flap angle correction of oswald efficiency factor
adel6 = np.array(
    [
        1.0,
        0.995,
        0.99,
        0.98,
        0.97,
        0.955,
        0.935,
        0.90,
        0.875,
        0.855,
        0.83,
        0.80,
        0.70,
        0.54,
        0.30,
    ]
)
# induced drag correction factors
asigma = np.array(
    [
        0.0,
        0.16,
        0.285,
        0.375,
        0.435,
        0.48,
        0.52,
        0.55,
        0.575,
        0.58,
        0.59,
        0.60,
        0.62,
        0.635,
        0.65,
    ]
)


def deg2rad(d):
    """Complex step safe deg2rad"""
    return d * np.pi / 180.0


def rad2deg(r):
    """Complex step safe rad2deg"""
    return r * 180.0 / np.pi


def cla(ar, sweep, mach):
    """Lift-curve slope of 3D wings from Seckel equation

    Parameters
    ----------
    ar : float
        Aspect ratio
    sweep : float
        Quarter-chord sweep angle, in radians
    mach : float
        Mach number.
    """
    return (
        np.pi
        * ar
        / (
            1
            + np.sqrt(
                1
                + (
                    ((ar / (2 * np.cos(sweep))) ** 2)
                    * (1 - (mach * np.cos(sweep)) ** 2)
                )
            )
        )
    )


def sigmoid(x, x0, alpha=0.1):
    """Sigmoid used to smoothly transition between piecewise functions"""
    if alpha == 0:
        raise ValueError("alpha must be non-zero")
    return 1 / (1 + np.exp(-(x - x0) / alpha))


class WingTailRatios(om.ExplicitComponent):
    """Static calculation of ratios between tail and wing parameters"""

    def setup(self):

        add_aviary_input(self, Aircraft.Wing.AREA, val=1370.3)

        add_aviary_input(self, Aircraft.Wing.SPAN, val=0.0)

        add_aviary_input(self, Aircraft.Wing.AVERAGE_CHORD, val=0.0)

        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, val=0.33)

        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15)

        add_aviary_input(self, Aircraft.Wing.MOUNTING_TYPE, val=0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, val=0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.SPAN, val=0.0)

        add_aviary_input(self, Aircraft.VerticalTail.SPAN, val=0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.AREA, val=0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.AVERAGE_CHORD, val=0.0)

        add_aviary_input(self, Aircraft.Fuselage.AVG_DIAMETER, val=0.0)

        self.add_output(
            "hbar", val=0.0, units="unitless",
            desc="HBAR: Ratio of HGAP(?) to wing span")
        self.add_output(
            "bbar", units="unitless", desc="BBAR: Ratio of H tail area to wing area")
        self.add_output(
            "sbar", units="unitless", desc="SBAR: Ratio of H tail area to wing area")
        self.add_output(
            "cbar", units="unitless", desc="SBAR: Ratio of H tail chord to wing chord")

    def setup_partials(self):
        self.declare_partials(
            "hbar",
            [
                Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION,
                Aircraft.VerticalTail.SPAN,
                Aircraft.Fuselage.AVG_DIAMETER,
                Aircraft.Wing.MOUNTING_TYPE,
                Aircraft.Wing.THICKNESS_TO_CHORD_ROOT,
                Aircraft.Wing.AREA,
                Aircraft.Wing.SPAN,
                Aircraft.Wing.TAPER_RATIO,
            ],
            method="cs",
        )
        self.declare_partials(
            "bbar", [Aircraft.HorizontalTail.SPAN, Aircraft.Wing.SPAN], method="cs"
        )
        self.declare_partials(
            "sbar", [Aircraft.HorizontalTail.AREA, Aircraft.Wing.AREA], method="cs"
        )
        self.declare_partials(
            "cbar",
            [Aircraft.HorizontalTail.AVERAGE_CHORD, Aircraft.Wing.AVERAGE_CHORD],
            method="cs",
        )

    def compute(self, inputs, outputs):
        (
            wing_area,
            wingspan,
            avg_chord,
            taper_ratio,
            tc_ratio_root,
            wing_loc,
            htail_loc,
            span_htail,
            span_vtail,
            htail_area,
            htail_chord,
            cabin_width,
        ) = inputs.values()

        trtw = tc_ratio_root * 2 * wing_area / wingspan / (1 + taper_ratio)
        hgap = cs.abs(
            htail_loc * span_vtail - 0.5 * (cabin_width - trtw) * (2 * wing_loc - 1)
        )
        outputs["hbar"] = hgap / wingspan
        outputs["bbar"] = span_htail / wingspan
        outputs["sbar"] = htail_area / wing_area
        outputs["cbar"] = htail_chord / avg_chord


class Xlifts(om.ExplicitComponent):
    """Compute lift ratio and lift-curve slope for given stability margin"""

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)

    def setup(self):
        nn = self.options["num_nodes"]

        # dynamic inputs
        self.add_input(Dynamic.Mission.MACH, val=0.0, units="unitless",
                       shape=nn, desc="Mach number")

        # stability inputs

        add_aviary_input(self, Aircraft.Design.STATIC_MARGIN, val=0.03)

        add_aviary_input(self, Aircraft.Design.CG_DELTA, val=0.25)

        # geometry inputs

        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, val=10.13)

        add_aviary_input(self, Aircraft.Wing.SWEEP, val=25.0)

        add_aviary_input(self, Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, val=0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.SWEEP, val=25.0)

        add_aviary_input(self, Aircraft.HorizontalTail.MOMENT_RATIO, val=0.0)

        # geometry from wing-tail ratios
        self.add_input(
            "sbar", units="unitless", desc="SBAR: Ratio of H tail area to wing area")
        self.add_input(
            "cbar", units="unitless", desc="CBAR: Ratio of H tail chord to wing chord")
        self.add_input(
            "hbar", units="unitless", desc="HBAR: Ratio of HGAP(?) to wing span")
        self.add_input(
            "bbar", units="unitless", desc="BBAR: Ratio of H tail area to wing area")

        self.add_output("lift_curve_slope", units="unitless",
                        shape=nn, desc="Lift-curve slope")
        self.add_output("lift_ratio", units="unitless", shape=nn, desc="Lift ratio")

    def setup_partials(self):
        ar = np.arange(self.options["num_nodes"])

        self.declare_partials("lift_ratio", "*", method="cs")
        self.declare_partials("lift_ratio", Dynamic.Mission.MACH,
                              rows=ar, cols=ar, method="cs")
        self.declare_partials("lift_curve_slope", "*", method="cs")
        self.declare_partials(
            "lift_curve_slope",
            [
                Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Wing.SWEEP,
                Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION,
                Aircraft.HorizontalTail.SWEEP,
                Aircraft.HorizontalTail.MOMENT_RATIO,
                "sbar",
                "cbar",
                "hbar",
                "bbar",
            ],
            method="cs",
        )
        self.declare_partials("lift_curve_slope", Dynamic.Mission.MACH,
                              rows=ar, cols=ar, method="cs")

    def compute(self, inputs, outputs):
        (
            mach,
            static_margin,
            delta_cg,
            AR,
            sweep_c4,
            htail_loc,
            htail_sweep,
            h_tail_moment,
            sbar,
            cbar,
            hbar,
            bbar,
        ) = inputs.values()

        delta = (static_margin + delta_cg) * h_tail_moment

        # TODO handle xt < 0?
        xt = 1 / h_tail_moment

        art = AR * bbar**2 / sbar
        h = hbar * AR

        # stability contribution from each surface
        claw0 = cla(AR, deg2rad(sweep_c4), mach)
        clat0 = cla(art, deg2rad(htail_sweep), mach) * (0.9 + 0.1 * htail_loc)

        # Hayes reverse flow theorem to estimate downwash effects on wing and canard
        eps1 = 1 / (4 * np.pi * np.sqrt(xt**2 + h**2))
        eps2 = 1 / np.pi / AR
        eps3 = cs.abs(xt) / (np.pi * AR * np.sqrt(xt**2 + h**2 + AR**2 / 4))
        eps4 = 1 / np.pi / art
        eps5 = cs.abs(xt) / (
            np.pi * art * np.sqrt(xt**2 + h**2 + art**2 * cbar**2 / 4)
        )

        claw = (
            claw0
            * (1 - clat0 * (eps4 - eps5 - cbar * eps1))
            / (1 - clat0 * claw0 * (eps1 + eps2 + eps3) * (eps4 - eps5 - cbar * eps1))
        )

        clat = clat0 * (1 - claw * (eps1 + eps2 + eps3))

        abar = clat / claw
        c = 1 / (1 + 1 / abar / sbar)
        lift_ratio = (c - delta) / (1 + delta - c)

        outputs["lift_curve_slope"] = claw
        outputs["lift_ratio"] = lift_ratio


class AeroGeom(om.ExplicitComponent):
    """Compute drag parameters from cruise conditions and geometric parameters.

    This corresponds to the AERO subroutine in GASP. The primary outputs are parameters
    SA* which build up the total aircraft drag coefficient.
    """

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)
        self.options.declare(
            "include_strut",
            default=False,
            types=bool,
            desc="Whether the aircraft has a strut or not",
        )

    def setup(self):
        nn = self.options["num_nodes"]

        self.add_input(
            Dynamic.Mission.MACH, val=0.0, units="unitless", shape=nn, desc="Current Mach number")
        self.add_input(
            Dynamic.Mission.SPEED_OF_SOUND,
            val=1.0,
            units="ft/s",
            shape=nn,
            desc="Speed of sound at current altitude",
        )
        self.add_input(
            "nu",
            val=1.0,
            units="ft**2/s",
            shape=nn,
            desc="Kinematic viscosity at current altitude",
        )

        self.add_input("ufac", units="unitless", shape=nn, desc="UFAC")

        # form factors
        # user could input these directly or use functions to estimate from geometry

        add_aviary_input(self, Aircraft.Wing.FORM_FACTOR, val=1.25)

        add_aviary_input(self, Aircraft.Fuselage.FORM_FACTOR, val=1.25)

        add_aviary_input(self, Aircraft.Nacelle.FORM_FACTOR, val=1.5)

        add_aviary_input(self, Aircraft.VerticalTail.FORM_FACTOR, val=1.25)

        add_aviary_input(self, Aircraft.HorizontalTail.FORM_FACTOR, val=1.25)

        add_aviary_input(self, Aircraft.Wing.FUSELAGE_INTERFERENCE_FACTOR, val=1.1)

        add_aviary_input(self, Aircraft.Strut.FUSELAGE_INTERFERENCE_FACTOR, val=0.0)

        # miscellaneous top-level inputs

        add_aviary_input(self, Aircraft.Design.DRAG_COEFFICIENT_INCREMENT, val=0.00175)

        add_aviary_input(self, Aircraft.Fuselage.FLAT_PLATE_AREA_INCREMENT, val=0.25)

        add_aviary_input(self, Aircraft.Wing.CENTER_DISTANCE, val=0.463)

        add_aviary_input(self, Aircraft.Wing.MIN_PRESSURE_LOCATION, val=0.3)

        add_aviary_input(self, Aircraft.Wing.MAX_THICKNESS_LOCATION, val=0.4)

        # geometric user inputs

        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, val=10.13)

        add_aviary_input(self, Aircraft.Wing.SWEEP, val=25.0)

        add_aviary_input(self, Aircraft.Wing.MOUNTING_TYPE, val=0.0)

        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, val=0.33)

        add_aviary_input(self, Aircraft.Strut.AREA_RATIO, val=0.0)

        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15)

        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD_TIP, val=0.12)

        # geometric data from sizing

        add_aviary_input(self, Aircraft.Wing.SPAN, val=0.0)

        add_aviary_input(self, Aircraft.Wing.AVERAGE_CHORD, val=0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.AVERAGE_CHORD, val=0.0)

        add_aviary_input(self, Aircraft.VerticalTail.AVERAGE_CHORD, val=0.0)

        add_aviary_input(self, Aircraft.Fuselage.LENGTH, val=0.0)

        add_aviary_input(self, Aircraft.Nacelle.AVG_LENGTH, val=0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.AREA, val=0.0)

        add_aviary_input(self, Aircraft.Fuselage.WETTED_AREA, val=0.0)

        add_aviary_input(self, Aircraft.Nacelle.SURFACE_AREA, val=0.0)

        add_aviary_input(self, Aircraft.Wing.AREA, val=1370.3)

        add_aviary_input(self, Aircraft.Fuselage.AVG_DIAMETER, val=0.0)

        add_aviary_input(self, Aircraft.VerticalTail.AREA, val=0.0)

        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED, val=0.0)

        add_aviary_input(self, Aircraft.Strut.CHORD, val=0.0)

        # outputs
        for i in range(7):
            name = f"SA{i+1}"
            self.add_output(name, units="unitless", shape=nn, desc=f"{name}: Drag param")

        self.add_output(
            "cf", units="unitless", shape=nn,
            desc="CFIN: Skin friction coefficient at Re=1e7"
        )

    def setup_partials(self):
        # self.declare_coloring(method="cs", show_summary=False)
        self.declare_partials("*", "*", dependent=False)
        ar = np.arange(self.options["num_nodes"])

        self.declare_partials(
            "SA1",
            [
                Aircraft.Wing.MIN_PRESSURE_LOCATION,
                Aircraft.Wing.MAX_THICKNESS_LOCATION,
                Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Wing.SWEEP,
                Aircraft.Wing.TAPER_RATIO,
                Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED,
            ],
            method="cs",
        )
        self.declare_partials(
            "SA2",
            [
                Aircraft.Wing.MIN_PRESSURE_LOCATION,
                Aircraft.Wing.MAX_THICKNESS_LOCATION,
                Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Wing.SWEEP,
                Aircraft.Wing.TAPER_RATIO,
            ],
            method="cs",
        )
        self.declare_partials(
            "SA3",
            [
                Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Wing.SWEEP,
                Aircraft.Wing.TAPER_RATIO,
                Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED,
            ],
            method="cs",
        )
        self.declare_partials(
            "SA4", [Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED], method="cs"
        )
        self.declare_partials("cf", [Dynamic.Mission.MACH],
                              rows=ar, cols=ar, method="cs")

        # diag partials for SA5-SA7
        self.declare_partials(
            "SA5", [Dynamic.Mission.MACH, Dynamic.Mission.SPEED_OF_SOUND, "nu"], rows=ar, cols=ar, method="cs"
        )
        self.declare_partials(
            "SA6", [Dynamic.Mission.MACH, Dynamic.Mission.SPEED_OF_SOUND, "nu"], rows=ar, cols=ar, method="cs"
        )
        self.declare_partials(
            "SA7", [Dynamic.Mission.MACH, Dynamic.Mission.SPEED_OF_SOUND, "nu", "ufac"], rows=ar, cols=ar, method="cs"
        )

        # dense partials for SA5-SA7
        most_params = [
            Aircraft.Wing.FORM_FACTOR,
            Aircraft.Fuselage.FORM_FACTOR,
            Aircraft.Nacelle.FORM_FACTOR,
            Aircraft.VerticalTail.FORM_FACTOR,
            Aircraft.HorizontalTail.FORM_FACTOR,
            Aircraft.Wing.FUSELAGE_INTERFERENCE_FACTOR,
            Aircraft.Strut.FUSELAGE_INTERFERENCE_FACTOR,
            Aircraft.Design.DRAG_COEFFICIENT_INCREMENT,
            Aircraft.Fuselage.FLAT_PLATE_AREA_INCREMENT,
            Aircraft.Wing.CENTER_DISTANCE,
            Aircraft.Wing.MOUNTING_TYPE,
            Aircraft.Wing.TAPER_RATIO,
            Aircraft.Strut.AREA_RATIO,
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT,
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP,
            Aircraft.Wing.SPAN,
            Aircraft.Wing.AVERAGE_CHORD,
            Aircraft.HorizontalTail.AVERAGE_CHORD,
            Aircraft.VerticalTail.AVERAGE_CHORD,
            Aircraft.Fuselage.LENGTH,
            Aircraft.Nacelle.AVG_LENGTH,
            Aircraft.HorizontalTail.AREA,
            Aircraft.Fuselage.WETTED_AREA,
            Aircraft.Nacelle.SURFACE_AREA,
            Aircraft.Wing.AREA,
            Aircraft.Fuselage.AVG_DIAMETER,
            Aircraft.VerticalTail.AREA,
        ]
        self.declare_partials("SA5", most_params, method="cs")
        self.declare_partials(
            "SA6", [Aircraft.Wing.FORM_FACTOR, Aircraft.Wing.AVERAGE_CHORD], method="cs"
        )
        self.declare_partials(
            "SA7",
            most_params + [Aircraft.Wing.ASPECT_RATIO, Aircraft.Wing.SWEEP],
            method="cs",
        )

    def compute(self, inputs, outputs):
        (
            mach,
            sos,
            nu,
            ufac,
            ff_wing,
            ff_fus,
            ff_nac,
            ff_vtail,
            ff_htail,
            wing_fus_intf,
            strut_fus_intf,
            cd0_inc,
            fe_fus_inc,
            wing_center_dist,
            wing_min_pressure_loc,
            wing_max_thickness_loc,
            AR,
            sweep_c4,
            wing_loc,
            taper_ratio,
            strut_wing_area_ratio,
            tc_ratio_root,
            tc_ratio_tip,
            wingspan,
            avg_chord,
            htail_chord,
            vtail_chord,
            fus_len,
            nac_len,
            htail_area,
            fus_SA,
            nacelle_area,
            wing_area,
            cabin_width,
            vtail_area,
            tc_ratio,
            strut_chord,
        ) = inputs.values()
        # skin friction coeff at Re = 10**7
        cf = 0.455 / 7**2.58 / (1 + 0.144 * mach**2) ** 0.65

        t = cs.abs(np.tan(deg2rad(sweep_c4)))
        yale05 = (1 - taper_ratio) / (1 + taper_ratio)
        # sweep angle to min pressure point
        dlmps = rad2deg(
            cs.arctan2(AR * t - 4 * (wing_min_pressure_loc - 0.25) * yale05, AR)
        )
        # sweep angle to max thickness point
        dlmtcx = rad2deg(
            cs.arctan2(AR * t - 4 * (wing_max_thickness_loc - 0.25) * yale05, AR)
        )
        # sweep angle of the leading edge
        rlmle = cs.arctan2(AR * t + yale05, AR)
        fk = 1 / (1 + yale05 / AR * 4 * taper_ratio**2)

        # Reynolds number per foot
        # here we make a smooth transition between a minimum reli (approximately
        # corresponding to Mach 0.1 at SLS) to help with takeoff. GASP doesn't call AERO
        # before takeoff, so the RELI used corresponds to the cruise point, and this
        # isn't a problem.
        reli_y1 = 700000 * np.ones(self.options["num_nodes"])
        reli_y2 = sos * mach / nu
        sig = sigmoid(mach, 0.1, alpha=0.005)
        reli = (1 - sig) * reli_y1 + sig * reli_y2

        # Re correction factors: fuselage, wing, nacelle, vtail, htail, strut, tip tank
        # protect against Mach 0, any other small Mach should be ok
        dtype = complex if self.under_complex_step else float
        ffre, fwre, fnre, fvtre, fhtre, fstrtre = np.ones(
            (6, self.options["num_nodes"]), dtype=dtype
        )
        if self.under_complex_step:
            good_mask = reli.real > 1
        else:
            good_mask = reli > 1
        ffre[good_mask] = (np.log10(reli[good_mask] * fus_len) / 7) ** -2.6
        fwre[good_mask] = (np.log10(reli[good_mask] * avg_chord) / 7) ** -2.6
        fnre[good_mask] = (np.log10(reli[good_mask] * nac_len) / 7) ** -2.6
        fvtre[good_mask] = (np.log10(reli[good_mask] * vtail_chord) / 7) ** -2.6
        fhtre[good_mask] = (np.log10(reli[good_mask] * htail_chord) / 7) ** -2.6
        if self.options["include_strut"]:
            fstrtre = (np.log10(reli[good_mask] * strut_chord) / 7) ** -2.6

        # fuselage form drag factor
        fffus = (
            1 + 1.5 * (cabin_width / fus_len) ** 1.5 + 7 * (cabin_width / fus_len) ** 3
        )

        # flat plate equivalent areas
        fef = ff_fus * fus_SA * cf * ffre * fffus + fe_fus_inc
        few = ff_wing * wing_area * cf * fwre
        # TODO replace 2 with num_engines
        fen = 2 * ff_nac * nacelle_area * cf * fnre
        fevt = ff_vtail * vtail_area * cf * fvtre
        feht = ff_htail * htail_area * cf * fhtre
        festrt = strut_fus_intf * strut_wing_area_ratio * wing_area * cf * fstrtre

        # begin INTERFERENCE - get flat plate equivalent for wing-fuselage interference
        croot = 2 * wing_area / (wingspan * (1 + taper_ratio))
        # wing profile drag coefficient
        cdw0 = few / wing_area
        zw_rf = 2 * wing_loc - 1
        x = tc_ratio_root * croot / cabin_width
        if cs.abs(zw_rf + x) >= 1:
            widthftop = 0.0
        else:
            widthftop = cabin_width * np.sqrt(1 - (zw_rf + x) ** 2)
        if cs.abs(zw_rf - x) >= 1:
            widthfbot = 0.0
        else:
            widthfbot = cabin_width * np.sqrt(1 - (zw_rf - x) ** 2)
        wbodywf = 0.5 * (widthftop + widthfbot)
        tcbodywf = tc_ratio_root - wbodywf / wingspan * (tc_ratio_root - tc_ratio_tip)
        cbodywf = croot * (1 - wbodywf / wingspan * (1 - taper_ratio))
        # factor due to vertical location
        kvwf = ckv[0] + zw_rf * (ckv[1] + zw_rf * ckv[2])
        # factor due to longitudinal location
        klwf = ckl[0] + wing_center_dist * (
            ckl[1]
            + wing_center_dist
            * (ckl[2] + wing_center_dist * (ckl[3] + wing_center_dist * ckl[4]))
        )
        # factor due to fuselage diameter / thickness
        kdtwf = ckdt[0] + ckdt[1] * cabin_width / (tcbodywf * cbodywf)
        # interference drag independent of shielded area
        feintwf = 1.5 * tcbodywf**3 * cbodywf**2 * kvwf * klwf * kdtwf
        areashieldwf = 0.5 * (croot + cbodywf) * wbodywf
        feshieldwf = cdw0 * areashieldwf
        feiwf = wing_fus_intf * (feintwf - feshieldwf)
        # end INTERFERENCE

        # total flat plate equivalent area
        fe = few + fef + fevt + feht + fen + feiwf + festrt + cd0_inc * wing_area

        wfob = cabin_width / wingspan
        siwb = (
            1
            - 0.0088 * wfob
            - 1.7364 * wfob**2
            - 2.303 * wfob**3
            + 6.0606 * wfob**4
        )

        # wing-free profile drag coefficient
        cdpo = (fe - few) / wing_area

        # Oswald efficiency
        see = 1.0 / (
            (1 / ufac / siwb)
            + 1.1938 * AR * (cdw0 / (np.cos(deg2rad(sweep_c4))) ** 2 + cdpo)
        )

        # compressibility drag parameters
        # sa1--4 are static, depending only on geometry
        sa1 = (1 + 0.0033 * (4 * dlmps - 3 * dlmtcx)) * (
            1 - 1.4 * tc_ratio - 0.06 * (1 - wing_min_pressure_loc)
        ) - 0.0368
        sa2 = (
            -0.33
            * (0.65 - wing_min_pressure_loc)
            * (1 + 0.0033 * (4 * dlmps - 3 * dlmtcx))
        )
        sa3 = (1.5 - 2 * fk**2 * np.sin(rlmle) ** 2) * tc_ratio ** (5 / 3.0)
        sa4 = 0.75 * tc_ratio

        # profile drag of everything but the wing
        sa5 = cdpo
        # wing profile drag
        sa6 = ff_wing * fwre
        # induced drag
        sa7 = 1.0 / (np.pi * see * AR)

        outputs["SA1"] = sa1
        outputs["SA2"] = sa2
        outputs["SA3"] = sa3
        outputs["SA4"] = sa4
        outputs["SA5"] = sa5
        outputs["SA6"] = sa6
        outputs["SA7"] = sa7
        outputs["cf"] = cf


class AeroSetup(om.Group):
    """Calculations for setting up aero"""

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)
        self.options.declare(
            "input_atmos",
            default=False,
            types=bool,
            desc="Directly input speed of sound and kinematic viscosity instead of "
            "computing them with an atmospherics component. For testing.",
        )

    def setup(self):
        nn = self.options["num_nodes"]

        self.add_subsystem("ratios", WingTailRatios(), promotes=["*"])
        self.add_subsystem("xlifts", Xlifts(num_nodes=nn), promotes=["*"])

        # implements EAERO
        interp = om.MetaModelStructuredComp(method="2D-slinear")
        interp.add_input("bbar", 0.0, units="unitless", training_data=xbbar)
        interp.add_input("hbar", 0.0, units="unitless", training_data=xhbar)
        interp.add_output("sigma", 0.0, units="unitless", training_data=sig1)
        interp.add_output("sigstr", 0.0, units="unitless", training_data=sig2)
        self.add_subsystem("interp", interp, promotes=["*"])

        self.add_subsystem(
            "ufac_calc",
            om.ExecComp(
                "ufac=(1 + lift_ratio)**2 / (sigstr*(lift_ratio/bbar)**2 + 2*sigma*lift_ratio/bbar + 1)",
                lift_ratio={'units': "unitless", "shape": nn},
                bbar={'units': "unitless"},
                sigma={'units': "unitless"},
                sigstr={'units': "unitless"},
                ufac={'units': "unitless", "shape": nn},
                has_diag_partials=True
            ),
            promotes=["*"],
        )

        if not self.options["input_atmos"]:
            # self.add_subsystem(
            #     "atmos",
            #     USatm1976Comp(num_nodes=nn),
            #     promotes_inputs=[("h", Dynamic.Mission.ALTITUDE)],
            #     promotes_outputs=["rho", Dynamic.Mission.SPEED_OF_SOUND, "viscosity"],
            # )
            self.add_subsystem(
                "kin_visc",
                om.ExecComp(
                    "nu = viscosity / rho",
                    viscosity={"units": "lbf*s/ft**2", "shape": nn},
                    rho={"units": "slug/ft**3", "shape": nn},
                    nu={"units": "ft**2/s", "shape": nn},
                    has_diag_partials=True
                ),
                promotes=["*"],
            )

        self.add_subsystem("geom", AeroGeom(num_nodes=nn), promotes=["*"])


class DragCoef(om.ExplicitComponent):
    """GASP lift coefficient calculation for low-speed near-ground flight.

    Drag for low-speed flight is affected by ground effects, landing gear, and flap
    deflection. Flaps are treated separately as an increment so their effects may be
    gradually added/removed over time.
    """

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)

    def setup(self):
        nn = self.options["num_nodes"]

        # dynamic inputs
        self.add_input(Dynamic.Mission.ALTITUDE, val=0.0,
                       units="ft", shape=nn, desc="Altitude")
        self.add_input(
            "CL", val=1.0, units="unitless", shape=nn, desc="Lift coefficient")

        # user inputs

        add_aviary_input(self, Mission.Design.GROSS_MASS, val=175400.0)

        self.add_input("flap_defl", val=10.0, units="deg", desc="Full flap deflection")

        add_aviary_input(self, Aircraft.Wing.HEIGHT, val=8.0)

        self.add_input(
            "airport_alt", val=0.0, units="ft", desc="HPORT: Airport altitude"
        )

        add_aviary_input(self, Aircraft.Wing.FLAP_CHORD_RATIO, val=0.3)

        # from flaps
        self.add_input(
            "dCL_flaps_model", val=0.0, units="unitless",
            desc="Delta CL from flaps model")
        self.add_input(
            "dCD_flaps_model", val=0.0, units="unitless",
            desc="Delta CD from flaps model")
        self.add_input(
            "dCL_flaps_coef",
            val=1.0, units="unitless",
            desc="SIGMTO | SIGMLD: Coefficient applied to delta CL from flaps model",
        )
        self.add_input(
            "CDI_factor",
            val=1.0, units="unitless",
            desc="VDEL6T | VDEL6L: Factor applied to induced drag with flaps",
        )

        # from sizing

        add_aviary_input(self, Aircraft.Wing.AVERAGE_CHORD, val=0.0)

        add_aviary_input(self, Aircraft.Wing.SPAN, val=0.0)

        add_aviary_input(self, Aircraft.Wing.AREA, val=1370.3)

        # from aero setup
        self.add_input(
            "cf", units="unitless", shape=nn,
            desc="CFIN: Skin friction coefficient at Re=1e7")
        self.add_input("SA5", units="unitless", shape=nn, desc="SA5: Drag param")
        self.add_input("SA6", units="unitless", shape=nn, desc="SA6: Drag param")
        self.add_input("SA7", units="unitless", shape=nn, desc="SA7: Drag param")

        self.add_output("CD_base", units="unitless", shape=nn, desc="Drag coefficient")
        self.add_output(
            "dCD_flaps_full", units="unitless", shape=nn,
            desc="CD increment with full flap deflection")
        self.add_output(
            "dCD_gear_full", units="unitless",
            shape=nn, desc="CD increment with landing gear down")

    def setup_partials(self):
        # self.declare_coloring(method="cs", show_summary=False)
        self.declare_partials("*", "*", dependent=False)
        ar = np.arange(self.options["num_nodes"])

        self.declare_partials("CD_base", ["*"], method="cs")
        self.declare_partials(
            "CD_base",
            [Dynamic.Mission.ALTITUDE, "CL", "cf", "SA5", "SA6", "SA7"],
            rows=ar,
            cols=ar,
            method="cs",
        )
        # self.declare_partials(
        #     "CD_base", [Mission.Design.GROSS_MASS, "dCD_flaps_model", "wing_area"], val=0
        # )

        self.declare_partials("dCD_flaps_full", ["dCD_flaps_model"], val=1)

        self.declare_partials(
            "dCD_gear_full",
            [Mission.Design.GROSS_MASS, Aircraft.Wing.AREA, "flap_defl"],
            method="cs",
        )

    def compute(self, inputs, outputs):
        (
            alt,
            CL,
            gross_mass_initial,
            flap_defl,
            wing_height,
            airport_alt,
            flap_chord_ratio,
            dCL_flaps_model,
            dCD_flaps_model,
            dCL_flaps_coef,
            CDI_factor,
            avg_chord,
            wingspan,
            wing_area,
            cf,
            SA5,
            SA6,
            SA7,
        ) = inputs.values()
        gross_wt_initial = gross_mass_initial * GRAV_ENGLISH_LBM

        # profile drag
        cd0 = SA5 + SA6 * cf

        # induced drag
        cdi = SA7 * (CL - dCL_flaps_coef * dCL_flaps_model) ** 2 / CDI_factor

        # ground effects - direct increment to CD
        hac = wing_height + alt - airport_alt
        heff = 2 * hac - np.sin(deg2rad(flap_defl)) * flap_chord_ratio * avg_chord
        sig = np.exp(-2.48 * (heff / wingspan) ** 0.768)
        betag = np.sqrt(1 + (heff / wingspan) ** 2) - heff / wingspan
        c1 = betag * CL / (12.5664 * hac)
        dcd_ground = -(sig - c1) * cdi / (1.0 - c1) - c1 * SA6 * cf

        # landing gear
        grfe = 0.0033 * gross_wt_initial**0.785
        dcd_gear = (grfe / wing_area) * (1 - 0.454545 * flap_defl / 50)

        outputs["CD_base"] = cd0 + cdi + dcd_ground
        outputs["dCD_flaps_full"] = dCD_flaps_model
        outputs["dCD_gear_full"] = dcd_gear


class DragCoefClean(om.ExplicitComponent):
    """Clean drag coefficient for high-speed flight"""

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)

    def setup(self):
        nn = self.options["num_nodes"]

        # dynamic inputs
        self.add_input(Dynamic.Mission.MACH, val=0.0, units="unitless",
                       shape=nn, desc="Mach number")
        self.add_input(
            "CL", val=1.0, units="unitless", shape=nn, desc="Lift coefficient")

        # user inputs

        add_aviary_input(self, Aircraft.Design.SUPERCRITICAL_DIVERGENCE_SHIFT, val=0.033)

        # from aero setup
        self.add_input(
            "cf", units="unitless", shape=nn,
            desc="CFIN: Skin friction coefficient at Re=1e7")
        self.add_input("SA1", units="unitless", shape=nn, desc="SA1: Drag param")
        self.add_input("SA2", units="unitless", shape=nn, desc="SA2: Drag param")
        self.add_input("SA5", units="unitless", shape=nn, desc="SA5: Drag param")
        self.add_input("SA6", units="unitless", shape=nn, desc="SA6: Drag param")
        self.add_input("SA7", units="unitless", shape=nn, desc="SA7: Drag param")

        self.add_output("CD", units="unitless", shape=nn, desc="Drag coefficient")

    def setup_partials(self):
        ar = np.arange(self.options["num_nodes"])

        self.declare_partials(
            "CD",
            [Dynamic.Mission.MACH, "CL", "cf", "SA1", "SA2", "SA5", "SA6", "SA7"],
            rows=ar,
            cols=ar,
            method="cs",
        )
        self.declare_partials(
            "CD", [Aircraft.Design.SUPERCRITICAL_DIVERGENCE_SHIFT], method="cs"
        )

    def compute(self, inputs, outputs):
        mach, CL, div_drag_supercrit, cf, SA1, SA2, SA5, SA6, SA7 = inputs.values()

        mach_div = SA1 + SA2 * CL + div_drag_supercrit

        sig = sigmoid(mach, mach_div, alpha=0.005)
        delcdm = sig * (10 * (mach - mach_div) ** 3)

        # delcdm = np.zeros_like(mach)
        # mask = np.bitwise_and(mach >= mach_div, mach > 0.6)
        # delcdm[mask] = 10 * (mach[mask] - mach_div[mask]) ** 3

        # profile drag
        cd0 = SA5 + SA6 * cf
        # induced drag
        cdi = SA7 * CL**2

        outputs["CD"] = cd0 + cdi + delcdm


class LiftCoeff(om.ExplicitComponent):
    """GASP lift coefficient calculation for low-speed near-ground flight"""

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)

    def setup(self):
        nn = self.options["num_nodes"]

        # dynamic inputs
        self.add_input("alpha", val=0.0, units="deg", shape=nn, desc="Angle of attack")
        self.add_input(Dynamic.Mission.ALTITUDE, val=0.0,
                       units="ft", shape=nn, desc="Altitude")
        self.add_input("lift_curve_slope", units="unitless",
                       shape=nn, desc="Lift-curve slope")
        self.add_input("lift_ratio", units="unitless", shape=nn, desc="Lift ratio")

        # user inputs

        add_aviary_input(self, Aircraft.Wing.ZERO_LIFT_ANGLE, val=-1.2)

        add_aviary_input(self, Aircraft.Wing.SWEEP, val=25.0)

        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, val=10.13)

        add_aviary_input(self, Aircraft.Wing.HEIGHT, val=8.0)

        self.add_input(
            "airport_alt", val=0.0, units="ft", desc="HPORT: Airport altitude"
        )

        self.add_input("flap_defl", val=10.0, units="deg", desc="Full flap deflection")

        add_aviary_input(self, Aircraft.Wing.FLAP_CHORD_RATIO, val=0.3)

        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, val=0.33)

        # from flaps
        self.add_input(
            "CL_max_flaps", units="unitless",
            desc="CLMWTO | CLMWLD: Max lift coefficient from flaps model",
        )
        self.add_input(
            "dCL_flaps_model", val=0.0, units="unitless",
            desc="Delta CL from flaps model")

        # from sizing

        add_aviary_input(self, Aircraft.Wing.AVERAGE_CHORD, val=0.0)

        add_aviary_input(self, Aircraft.Wing.SPAN, val=0.0)

        self.add_output(
            "CL_base", units="unitless", shape=nn, desc="Base lift coefficient")
        self.add_output(
            "dCL_flaps_full", units="unitless", shape=nn,
            desc="CL increment with full flap deflection"
        )
        self.add_output(
            "alpha_stall", units="deg", shape=nn, desc="Stall angle of attack"
        )
        self.add_output(
            "CL_max", units="unitless", shape=nn, desc="Max lift coefficient")

    def setup_partials(self):
        # self.declare_coloring(method="cs", show_summary=False)
        self.declare_partials("*", "*", dependent=False)
        ar = np.arange(self.options["num_nodes"])

        dynvars = ["alpha", Dynamic.Mission.ALTITUDE, "lift_curve_slope", "lift_ratio"]

        self.declare_partials("CL_base", ["*"], method="cs")
        self.declare_partials("CL_base", dynvars, rows=ar, cols=ar, method="cs")

        self.declare_partials("dCL_flaps_full", ["dCL_flaps_model"], method="cs")
        self.declare_partials(
            "dCL_flaps_full", ["lift_ratio"], rows=ar, cols=ar, method="cs")

        self.declare_partials("alpha_stall", ["*"], method="cs")
        self.declare_partials("alpha_stall", dynvars, rows=ar, cols=ar, method="cs")

        self.declare_partials("CL_max", ["CL_max_flaps"], method="cs")
        self.declare_partials("CL_max", ["lift_ratio"], rows=ar, cols=ar, method="cs")

    def compute(self, inputs, outputs):
        (
            alpha,
            alt,
            lift_curve_slope,
            lift_ratio,
            alpha0,
            sweep_c4,
            AR,
            wing_height,
            airport_alt,
            flap_defl,
            flap_chord_ratio,
            taper_ratio,
            CL_max_flaps,
            dCL_flaps_model,
            avg_chord,
            wingspan,
        ) = inputs.values()

        # ground effects - factor on lift-curve slope
        hac = wing_height + alt - airport_alt
        heff = 2 * hac - np.sin(deg2rad(flap_defl)) * flap_chord_ratio * avg_chord
        sig = np.exp(-2.48 * (heff / wingspan) ** 0.768)
        betag = (1 + (heff / wingspan) ** 2) ** 0.5 - heff / wingspan
        rlmc2 = cs.arctan2(
            AR * np.tan(deg2rad(sweep_c4)) - ((1 - taper_ratio) / (1 + taper_ratio)), AR
        )
        c3 = 2 * np.cos(rlmc2) + np.sqrt(AR**2 + (2 * np.cos(rlmc2)) ** 2)
        c4 = betag / (12.5664 * hac / avg_chord)
        cloge = lift_curve_slope * deg2rad(alpha - alpha0) + dCL_flaps_model
        kclge = (
            1
            + sig
            - sig * AR * np.cos(rlmc2) / c3
            - c4 * (cloge - lift_curve_slope / (16 * hac / avg_chord))
        )
        kclge = np.clip(kclge, 1.0, None)

        outputs["CL_base"] = kclge * lift_curve_slope * \
            deg2rad(alpha - alpha0) * (1 + lift_ratio)
        outputs["dCL_flaps_full"] = dCL_flaps_model * (1 + lift_ratio)

        outputs["alpha_stall"] = (
            rad2deg((CL_max_flaps - dCL_flaps_model) /
                    (kclge * lift_curve_slope)) + alpha0
        )
        outputs["CL_max"] = CL_max_flaps * (1 + lift_ratio)


class LiftCoeffClean(om.ExplicitComponent):
    """Clean wing lift coefficient for high-speed flight"""

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)
        self.options.declare(
            "output_alpha",
            default=False,
            types=bool,
            desc="If True, output alpha for a given input CL",
        )

    def setup(self):
        nn = self.options["num_nodes"]

        if self.options["output_alpha"]:
            self.add_output(
                "alpha", val=0.0, units="deg", shape=nn, desc="Angle of attack"
            )
            self.add_input(
                "CL", val=1.0, units="unitless", shape=nn, desc="Lift coefficient")
        else:
            self.add_input(
                "alpha", val=0.0, units="deg", shape=nn, desc="Angle of attack"
            )
            self.add_output(
                "CL", val=1.0, units="unitless", shape=nn, desc="Lift coefficient")

        self.add_input("lift_curve_slope", units="unitless",
                       shape=nn, desc="Lift-curve slope")
        self.add_input("lift_ratio", units="unitless", shape=nn, desc="Lift ratio")

        add_aviary_input(self, Aircraft.Wing.ZERO_LIFT_ANGLE, val=-1.2)

        add_aviary_input(self, Mission.Design.LIFT_COEFFICIENT_MAX_FLAPS_UP, val=0.0)

        self.add_output("alpha_stall", shape=nn, desc="Stall angle of attack")
        self.add_output(
            "CL_max", units="unitless", shape=nn, desc="Max lift coefficient")

    def setup_partials(self):
        # self.declare_coloring(method="cs", show_summary=False)
        self.declare_partials("*", "*", dependent=False)
        ar = np.arange(self.options["num_nodes"])

        if self.options["output_alpha"]:
            self.declare_partials(
                "alpha", ["CL", "lift_ratio", "lift_curve_slope"], rows=ar, cols=ar, method="cs"
            )
            self.declare_partials("alpha", [Aircraft.Wing.ZERO_LIFT_ANGLE], method="cs")
        else:
            self.declare_partials(
                "CL", ["lift_curve_slope", "alpha", "lift_ratio"], rows=ar, cols=ar, method="cs"
            )
            self.declare_partials("CL", [Aircraft.Wing.ZERO_LIFT_ANGLE], method="cs")

        self.declare_partials(
            "alpha_stall", ["lift_curve_slope"], rows=ar, cols=ar, method="cs")
        self.declare_partials(
            "alpha_stall",
            [
                Mission.Design.LIFT_COEFFICIENT_MAX_FLAPS_UP,
                Aircraft.Wing.ZERO_LIFT_ANGLE,
            ],
            method="cs",
        )

        self.declare_partials("CL_max", ["lift_ratio"], rows=ar, cols=ar, method="cs")
        self.declare_partials(
            "CL_max", [Mission.Design.LIFT_COEFFICIENT_MAX_FLAPS_UP], method="cs"
        )

    def compute(self, inputs, outputs):
        _, lift_curve_slope, lift_ratio, alpha0, CL_max_flaps = inputs.values()
        if self.options["output_alpha"]:
            CL = inputs["CL"]
            clw = CL / (1 + lift_ratio)
            outputs["alpha"] = rad2deg(clw / lift_curve_slope) + alpha0
        else:
            alpha = inputs["alpha"]
            outputs["CL"] = lift_curve_slope * deg2rad(alpha - alpha0) * (1 + lift_ratio)

        outputs["alpha_stall"] = rad2deg(CL_max_flaps / lift_curve_slope) + alpha0
        outputs["CL_max"] = CL_max_flaps * (1 + lift_ratio)


class CruiseAero(om.Group):
    """Top-level aerodynamics group for cruise (no flaps, no landing gear)"""

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)
        self.options.declare(
            "output_alpha",
            default=False,
            types=bool,
            desc="If True, output alpha for a given input CL",
        )
        self.options.declare(
            "input_atmos",
            default=False,
            types=bool,
            desc="Directly input speed of sound and kinematic viscosity instead of "
            "computing them with an atmospherics component. For testing.",
        )

    def setup(self):
        nn = self.options["num_nodes"]
        self.add_subsystem(
            "aero_setup",
            AeroSetup(num_nodes=nn, input_atmos=self.options["input_atmos"]),
            promotes=["*"],
        )
        if self.options["output_alpha"]:
            # lift_req -> CL
            self.add_subsystem("lift2cl", CLFromLift(num_nodes=nn), promotes=["*"])
        self.add_subsystem(
            "lift_coef",
            LiftCoeffClean(output_alpha=self.options["output_alpha"], num_nodes=nn),
            promotes=["*"],
        )
        self.add_subsystem("drag_coef", DragCoefClean(num_nodes=nn), promotes=["*"])
        self.add_subsystem("forces", AeroForces(num_nodes=nn), promotes=["*"])


class LowSpeedAero(om.Group):
    """Top-level aerodynamics group for near-ground flight"""

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)
        self.options.declare(
            "retract_gear",
            default=True,
            types=bool,
            desc="True to start with gear landing gear down, False for reverse",
        )
        self.options.declare(
            "retract_flaps",
            default=True,
            types=bool,
            desc="True to start with flaps applied, False for reverse",
        )
        self.options.declare(
            "output_alpha",
            default=False,
            types=bool,
            desc="If True, output alpha for a given input CL",
        )
        self.options.declare(
            "input_atmos",
            default=False,
            types=bool,
            desc="Directly input speed of sound and kinematic viscosity instead of "
            "computing them with an atmospherics component. For testing.",
        )

    def setup(self):
        nn = self.options["num_nodes"]
        output_alpha = self.options["output_alpha"]

        self.add_subsystem(
            "aero_setup",
            AeroSetup(num_nodes=nn, input_atmos=self.options["input_atmos"]),
            promotes=["*"],
        )

        aero_ramps = TanhRampComp(time_units='s', num_nodes=nn)
        aero_ramps.add_ramp('flap_factor', output_units='unitless',
                            initial_val=1.0 if self.options['retract_flaps'] else 0.0,
                            final_val=0.0 if self.options['retract_flaps'] else 1.0)
        aero_ramps.add_ramp('gear_factor', output_units='unitless',
                            initial_val=1.0 if self.options['retract_gear'] else 0.0,
                            final_val=0.0 if self.options['retract_gear'] else 1.0)

        self.add_subsystem("aero_ramps",
                           aero_ramps,
                           promotes_inputs=[("time", "t_curr"),
                                            ("flap_factor:t_init", "t_init_flaps"),
                                            ("flap_factor:t_duration", "dt_flaps"),
                                            ("gear_factor:t_init", "t_init_gear"),
                                            ("gear_factor:t_duration", "dt_gear")],
                           promotes_outputs=['flap_factor',
                                             'gear_factor'])

        if output_alpha:
            # lift_req -> CL
            self.add_subsystem(
                "lift2cl",
                CLFromLift(num_nodes=nn),
                promotes_inputs=["*"],
                # little bit of a hack here - input CL bypasses CL increment ramp
                # so ensure this is what's passed to DragCoef
                promotes_outputs=[("CL", "CL_full_flaps")],
            )
            # Note: alpha is not an output
        else:
            # TODO this makes "output_alpha" a bit of a misnomer, would computing an
            # alpha be useful? an issue is ground effects depend on alpha
            self.add_subsystem(
                "lift_coef",
                LiftCoeff(num_nodes=nn),
                promotes_inputs=["*"],
                promotes_outputs=["*"]
            )

            self.add_subsystem(
                "total_cl",
                om.ExecComp(
                    [
                        # "CL = CL_base + dCL_flaps",
                        "CL_full_flaps = CL_base + dCL_flaps_full",
                        "CL = CL_base + flap_factor * dCL_flaps_full",
                    ],
                    CL=dict(shape=nn, units='unitless'),
                    CL_full_flaps=dict(shape=nn, units='unitless'),
                    CL_base=dict(shape=nn, units='unitless'),
                    # dCL_flaps=dict(shape=nn, units='unitless'),
                    flap_factor=dict(shape=nn, units='unitless'),
                    dCL_flaps_full=dict(shape=nn, units='unitless'),
                    has_diag_partials=True
                ),
                promotes=["*"],
            )

        interp = om.MetaModelStructuredComp(method="slinear")
        interp.add_input("flap_defl", 10.0, units="deg", training_data=adelfd)
        interp.add_output("dCL_flaps_coef", 0.0, units="unitless", training_data=asigma)
        interp.add_output("CDI_factor", 1.0, units="unitless", training_data=adel6)
        self.add_subsystem("cdi_flap_interp", interp, promotes=["*"])

        self.add_subsystem(
            "drag_coef", DragCoef(num_nodes=nn), promotes=["*", ("CL", "CL_full_flaps")]
        )

        self.add_subsystem(
            "total_cd",
            om.ExecComp(
                "CD = CD_base + flap_factor * dCD_flaps_full + gear_factor"
                " * dCD_gear_full",
                # "CD = CD_base + dCD_flaps + dCD_gear",
                CD=dict(shape=nn, units='unitless'),
                CD_base=dict(shape=nn, units='unitless'),
                # dCD_flaps=dict(shape=nn, units='unitless'),
                # dCD_gear=dict(shape=nn, units='unitless'),
                flap_factor=dict(shape=nn, units='unitless'),
                gear_factor=dict(shape=nn, units='unitless'),
                dCD_gear_full=dict(shape=nn, units='unitless'),
                dCD_flaps_full=dict(shape=nn, units='unitless'),
                has_diag_partials=True
            ),
            promotes=["*"],
        )

        self.add_subsystem("forces", AeroForces(num_nodes=nn), promotes=["*"])

        self.set_input_defaults(Dynamic.Mission.ALTITUDE, np.zeros(nn))

        if self.options["retract_gear"]:
            # takeoff defaults
            self.set_input_defaults("dt_gear", 7)
        # gear not dynamically extended during landing

        if self.options["retract_flaps"]:
            # takeoff defaults
            self.set_input_defaults("flap_defl", 10)
            self.set_input_defaults("dt_flaps", 3)
        else:
            # landing defaults
            self.set_input_defaults("flap_defl", 40)
            # flaps not dynamically extended during landing



import unittest

import dymos as dm
import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal

from aviary.subsystems.propulsion.engine_model import EngineModel
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Dynamic
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs


class PreMissionEngine(om.Group):
    def setup(self):
        self.add_subsystem('dummy_comp', om.ExecComp(
            'y=x**2', x={'units': 'm', 'val': 2.}, y={'units': 'm**2'}), promotes=['*'])


class SimpleEngine(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']
        # add inputs and outputs to interpolator
        self.add_input(Dynamic.Mission.MACH,
                       shape=nn,
                       units='unitless',
                       desc='Current flight Mach number')
        self.add_input(Dynamic.Mission.ALTITUDE,
                       shape=nn,
                       units='ft',
                       desc='Current flight altitude')
        self.add_input(Dynamic.Mission.THROTTLE,
                       shape=nn,
                       units='unitless',
                       desc='Current engine throttle')
        self.add_input('different_throttle',
                       shape=nn,
                       units='unitless',
                       desc='Little bonus throttle for testing')
        self.add_input('y',
                       units='m**2',
                       desc='Dummy variable for bus testing')

        self.add_output(Dynamic.Mission.THRUST,
                        shape=nn,
                        units='lbf',
                        desc='Current net thrust produced (scaled)')
        self.add_output(Dynamic.Mission.THRUST_MAX,
                        shape=nn,
                        units='lbf',
                        desc='Current net thrust produced (scaled)')
        self.add_output(Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE,
                        shape=nn,
                        units='lbm/s',
                        desc='Current fuel flow rate (scaled)')
        self.add_output(Dynamic.Mission.ELECTRIC_POWER,
                        shape=nn,
                        units='W',
                        desc='Current electric energy rate (scaled)')
        self.add_output(Dynamic.Mission.NOX_RATE,
                        shape=nn,
                        units='lbm/s',
                        desc='Current NOx emission rate (scaled)')
        self.add_output(Dynamic.Mission.TEMPERATURE_ENGINE_T4,
                        shape=nn,
                        units='degR',
                        desc='Current turbine exit temperature')

        self.declare_partials('*', '*', method='fd')

    def compute(self, inputs, outputs):
        combined_throttle = inputs[Dynamic.Mission.THROTTLE] + \
            inputs['different_throttle']

        # calculate outputs
        outputs[Dynamic.Mission.THRUST] = 10000. * combined_throttle
        outputs[Dynamic.Mission.THRUST_MAX] = 10000.
        outputs[Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE] = -10. * combined_throttle
        outputs[Dynamic.Mission.TEMPERATURE_ENGINE_T4] = 2800.


class SimpleTestEngine(EngineModel):
    def __init__(self, name='engine', options=None):
        aviary_inputs = AviaryValues()
        super().__init__(name, options=aviary_inputs)

    def build_pre_mission(self, aviary_inputs=AviaryValues()):
        return PreMissionEngine()

    def build_mission(self, num_nodes, aviary_inputs):
        return SimpleEngine(num_nodes=num_nodes)

    def get_controls(self):
        controls_dict = {
            "different_throttle": {'units': 'unitless', 'lower': 0., 'upper': 0.1},
        }
        return controls_dict

    def get_bus_variables(self):
        bus_dict = {
            "y": {
                "mission_name": "y",
                "units": "m**2",
            },
        }
        return bus_dict

    def get_initial_guesses(self):
        initial_guesses_dict = {
            "different_throttle": {
                "val": 0.05,
                "units": "unitless",
                "type": "control",
            }
        }
        return initial_guesses_dict


@use_tempdirs
class CustomEngineTest(unittest.TestCase):
    def test_custom_engine(self):
        from aviary.interface.methods_for_level2 import AviaryProblem

        phase_info = {
            'pre_mission': {
                'include_takeoff': False,
                'external_subsystems': [],
                'optimize_mass': True,
            },
            'cruise': {
                "subsystem_options": {"core_aerodynamics": {"method": "computed"}},
                "user_options": {
                    "optimize_mach": False,
                    "optimize_altitude": False,
                    "polynomial_control_order": 1,
                    "num_segments": 2,
                    "order": 3,
                    "solve_for_distance": False,
                    "initial_mach": (0.72, "unitless"),
                    "final_mach": (0.72, "unitless"),
                    "mach_bounds": ((0.7, 0.74), "unitless"),
                    "initial_altitude": (35000.0, "ft"),
                    "final_altitude": (35000.0, "ft"),
                    "altitude_bounds": ((23000.0, 38000.0), "ft"),
                    "throttle_enforcement": "boundary_constraint",
                    "fix_initial": False,
                    "constrain_final": False,
                    "fix_duration": False,
                    "initial_bounds": ((0.0, 0.0), "min"),
                    "duration_bounds": ((10., 30.), "min"),
                },
                "initial_guesses": {"times": ([0, 30], "min")},
            },
            'post_mission': {
                'include_landing': False,
                'external_subsystems': [],
            }
        }

        prob = AviaryProblem(reports=False)

        # Load aircraft and options data from user
        # Allow for user overrides here
        prob.load_inputs("models/test_aircraft/aircraft_for_bench_GwFm.csv",
                         phase_info, engine_builder=SimpleTestEngine())

        # Preprocess inputs
        prob.check_and_preprocess_inputs()

        prob.add_pre_mission_systems()

        prob.add_phases()

        prob.add_post_mission_systems()

        # Link phases and variables
        prob.link_phases()

        prob.add_driver("SLSQP")

        prob.add_design_variables()

        prob.add_objective('fuel_burned')

        prob.setup()

        prob.set_initial_guesses()

        prob.final_setup()

        # check that the different throttle initial guess has been set correctly
        initial_guesses = prob.get_val(
            'traj.phases.cruise.controls:different_throttle')[0]
        assert_near_equal(float(initial_guesses), 0.05)

        # and run mission, and dynamics
        dm.run_problem(prob, run_driver=True, simulate=False, make_plots=False)

        tol = 1.e-4

        assert_near_equal(float(prob.get_val('traj.cruise.rhs_all.y')), 4., tol)


if __name__ == '__main__':
    unittest.main()



'''
Define utilities to calculate the estimated mass of passenger service
equipment.
'''
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class PassengerServiceMass(om.ExplicitComponent):
    '''
    Define the default component to calculate the estimated mass of passenger
    service equipment. The methodology is based on the
    FLOPS weight equations, modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(
            self,
            Aircraft.CrewPayload.PASSENGER_SERVICE_MASS_SCALER,
            val=1.,
        )

        add_aviary_input(
            self,
            Mission.Design.RANGE,
            val=0.0,
        )

        add_aviary_output(
            self,
            Aircraft.CrewPayload.PASSENGER_SERVICE_MASS,
            val=0.0,
        )

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        first_class_count = aviary_options.get_val(Aircraft.CrewPayload.NUM_FIRST_CLASS)

        business_class_count = \
            aviary_options.get_val(Aircraft.CrewPayload.NUM_BUSINESS_CLASS)

        tourist_class_count = \
            aviary_options.get_val(Aircraft.CrewPayload.NUM_TOURIST_CLASS)
        design_range = inputs[Mission.Design.RANGE]
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)

        passenger_service_mass_scaler = \
            inputs[Aircraft.CrewPayload.PASSENGER_SERVICE_MASS_SCALER]

        passenger_service_weight = (
            (
                5.164 * first_class_count
                + 3.846 * business_class_count
                + 2.529 * tourist_class_count
            ) * (design_range / max_mach)**0.225
        ) * passenger_service_mass_scaler

        outputs[Aircraft.CrewPayload.PASSENGER_SERVICE_MASS] = \
            passenger_service_weight / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J, discrete_inputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        first_class_count = aviary_options.get_val(Aircraft.CrewPayload.NUM_FIRST_CLASS)

        business_class_count = \
            aviary_options.get_val(Aircraft.CrewPayload.NUM_BUSINESS_CLASS)

        tourist_class_count = \
            aviary_options.get_val(Aircraft.CrewPayload.NUM_TOURIST_CLASS)
        design_range = inputs[Mission.Design.RANGE]
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)

        passenger_service_mass_scaler = \
            inputs[Aircraft.CrewPayload.PASSENGER_SERVICE_MASS_SCALER]

        J[
            Aircraft.CrewPayload.PASSENGER_SERVICE_MASS,
            Mission.Design.RANGE
        ] = passenger_service_mass_scaler * (
            5.164 * first_class_count
            + 3.846 * business_class_count
            + 2.529 * tourist_class_count
        ) * 0.225 * ((design_range / max_mach)**-0.775) / max_mach / GRAV_ENGLISH_LBM

        J[
            Aircraft.CrewPayload.PASSENGER_SERVICE_MASS,
            Aircraft.CrewPayload.PASSENGER_SERVICE_MASS_SCALER,
        ] = (
            (
                5.164 * first_class_count
                + 3.846 * business_class_count
                + 2.529 * tourist_class_count
            ) * (design_range / max_mach)**0.225
        ) / GRAV_ENGLISH_LBM


class AltPassengerServiceMass(om.ExplicitComponent):
    '''
    Define the alternate component to calculate the estimated mass of
    passenger service equipment. The methodology is based on the
    FLOPS weight equations, modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(
            self, Aircraft.CrewPayload.PASSENGER_SERVICE_MASS_SCALER, val=1.)

        add_aviary_output(self, Aircraft.CrewPayload.PASSENGER_SERVICE_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        aviary_options: AviaryValues = self.options['aviary_options']
        passenger_count = aviary_options.get_val(
            Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')

        passenger_service_mass_scaler = \
            inputs[Aircraft.CrewPayload.PASSENGER_SERVICE_MASS_SCALER]

        passenger_service_weight = \
            31.7 * passenger_count * passenger_service_mass_scaler

        outputs[Aircraft.CrewPayload.PASSENGER_SERVICE_MASS] = \
            passenger_service_weight / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J, discrete_inputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        passenger_count = aviary_options.get_val(
            Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')

        J[
            Aircraft.CrewPayload.PASSENGER_SERVICE_MASS,
            Aircraft.CrewPayload.PASSENGER_SERVICE_MASS_SCALER
        ] = 31.7 * passenger_count / GRAV_ENGLISH_LBM



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.mass.flops_based.distributed_prop import (
    distributed_nacelle_diam_factor, distributed_nacelle_diam_factor_deriv)
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission

DEG2RAD = np.pi / 180.0


class LandingGearMass(om.ExplicitComponent):
    '''
    Calculate the mass of the landing gear. The methodology is based on the
    FLOPS weight equations, modified to output mass instead of weight.
    '''
    # TODO: add in aircraft type and carrier factors as options and modify
    # equations

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH, val=0.0)

        add_aviary_input(self, Aircraft.LandingGear.MAIN_GEAR_MASS_SCALER, val=1.0)

        add_aviary_input(self, Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH, val=0.0)

        add_aviary_input(self, Aircraft.LandingGear.NOSE_GEAR_MASS_SCALER, val=1.0)

        add_aviary_input(self, Aircraft.Design.TOUCHDOWN_MASS, val=0.0)

        add_aviary_output(self, Aircraft.LandingGear.MAIN_GEAR_MASS, val=0.0)

        add_aviary_output(self, Aircraft.LandingGear.NOSE_GEAR_MASS, val=0.0)

        # TODO landing weight is not a landing_gear component level variable
        # self.add_input('aircraft:landing_gear:weights:landing_weight', val=0.0, desc='design landing weight', units='lbf')
        # self.add_input('aircraft:landing_gear:dimensions:extend_main_gear_oleo_len', val=0.0, desc='length of extended \
        #     main landing gear oleo', units='inch')
        # self.add_input('aircraft:landing_gear:dimensions:extend_nose_gear_oleo_len', val=0.0, desc='length of extended \
        #     nose landing gear oleo', units='inch')
        # self.add_input('TBD:aircraft:landing_gear:main_landing_gear_weight_multipler', val=1.0, desc='weight multiplier for \
        #     main landing gear weight', units='unitless')
        # self.add_input('TBD:aircraft:landing_gear:nose_landing_gear_weight_multipler', val=1.0, desc='weight multiplier for \
        #     nose landing gear weight', units='unitless')

        # self.add_output('TBD:landing_gear:weights:main_landing_gear_weight', val=0.0, desc='main landing gear weight', units='lbf')
        # self.add_output('TBD:landing_gear:weights:nose_landing_gear_weight', val=0.0, desc='nose landing gear weight', units='lbf')

    def setup_partials(self):
        self.declare_partials(
            Aircraft.LandingGear.MAIN_GEAR_MASS,
            [
                Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                Aircraft.LandingGear.MAIN_GEAR_MASS_SCALER,
                Aircraft.Design.TOUCHDOWN_MASS])
        self.declare_partials(
            Aircraft.LandingGear.NOSE_GEAR_MASS,
            [
                Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH,
                Aircraft.LandingGear.NOSE_GEAR_MASS_SCALER,
                Aircraft.Design.TOUCHDOWN_MASS])

    def compute(self, inputs, outputs):

        main_gear_length = inputs[Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH]
        main_gear_scaler = inputs[Aircraft.LandingGear.MAIN_GEAR_MASS_SCALER]
        nose_gear_length = inputs[Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH]
        nose_gear_scaler = inputs[Aircraft.LandingGear.NOSE_GEAR_MASS_SCALER]
        landing_weight = inputs[Aircraft.Design.TOUCHDOWN_MASS] * \
            GRAV_ENGLISH_LBM

        outputs[Aircraft.LandingGear.MAIN_GEAR_MASS] = 0.0117 * \
            landing_weight**0.95 * main_gear_length**0.43 * \
            main_gear_scaler / GRAV_ENGLISH_LBM
        outputs[Aircraft.LandingGear.NOSE_GEAR_MASS] = 0.048 * \
            landing_weight**0.67 * nose_gear_length**0.43 * \
            nose_gear_scaler / GRAV_ENGLISH_LBM

        # main_gear_weight = (0.0117 - 0.0012 * type_factor) * landing_weight**0.95 * main_gear_length**0.43
        # outputs['TBD:landing_gear:weights:main_landing_gear_weight'] = main_gear_weight * inputs['TBD:aircraft:landing_gear:main_landing_gear_weight_multipler']
        # nose_gear_weight = (0.048 - 0.008 * type_factor) * landing_weight**0.67 * nose_gear_length**0.43 * (1 + 0.8*carrier_factor)
        # outputs['TBD:landing_gear:weights:nose_landing_gear_weight'] = nose_gear_weight * inputs['TBD:aircraft:landing_gear:nose_landing_gear_weight_multipler']

    def compute_partials(self, inputs, J):
        main_gear_length = inputs[Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH]
        main_gear_scaler = inputs[Aircraft.LandingGear.MAIN_GEAR_MASS_SCALER]
        nose_gear_length = inputs[Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH]
        nose_gear_scaler = inputs[Aircraft.LandingGear.NOSE_GEAR_MASS_SCALER]
        landing_weight = \
            inputs[Aircraft.Design.TOUCHDOWN_MASS] * GRAV_ENGLISH_LBM

        landing_weight_exp1 = landing_weight**0.95
        landing_weight_exp2 = landing_weight**0.67
        main_gear_length_exp = main_gear_length**0.43
        nose_gear_length_exp = nose_gear_length**0.43

        J[
            Aircraft.LandingGear.MAIN_GEAR_MASS,
            Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH] = 0.005031 * \
            landing_weight_exp1 * main_gear_length**-0.57 * \
            main_gear_scaler / GRAV_ENGLISH_LBM
        J[
            Aircraft.LandingGear.MAIN_GEAR_MASS,
            Aircraft.LandingGear.MAIN_GEAR_MASS_SCALER] = 0.0117 * \
            landing_weight_exp1 * main_gear_length_exp / GRAV_ENGLISH_LBM
        J[Aircraft.LandingGear.MAIN_GEAR_MASS, Aircraft.Design.TOUCHDOWN_MASS] = \
            0.011115 * \
            landing_weight**-0.05 * main_gear_length_exp * main_gear_scaler

        J[
            Aircraft.LandingGear.NOSE_GEAR_MASS,
            Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH] = 0.02064 * \
            landing_weight_exp2 * nose_gear_length**-0.57 * \
            nose_gear_scaler / GRAV_ENGLISH_LBM
        J[
            Aircraft.LandingGear.NOSE_GEAR_MASS,
            Aircraft.LandingGear.NOSE_GEAR_MASS_SCALER] = 0.048 * \
            landing_weight_exp2 * nose_gear_length_exp / GRAV_ENGLISH_LBM
        J[Aircraft.LandingGear.NOSE_GEAR_MASS, Aircraft.Design.TOUCHDOWN_MASS] = \
            0.03216 * \
            landing_weight**-0.33 * nose_gear_length_exp * \
            nose_gear_scaler


class AltLandingGearMass(om.ExplicitComponent):
    '''
    Calculate the mass of the landing gear using the alternate method.
    The methodology is based on the FLOPS weight equations, modified
    to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH, val=0.0)

        add_aviary_input(self, Aircraft.LandingGear.MAIN_GEAR_MASS_SCALER, val=1.0)

        add_aviary_input(self, Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH, val=0.0)

        add_aviary_input(self, Aircraft.LandingGear.NOSE_GEAR_MASS_SCALER, val=1.0)

        add_aviary_input(self, Mission.Design.GROSS_MASS, val=0.0)

        add_aviary_output(self, Aircraft.LandingGear.MAIN_GEAR_MASS, val=0.0)

        add_aviary_output(self, Aircraft.LandingGear.NOSE_GEAR_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials(Aircraft.LandingGear.MAIN_GEAR_MASS, [
            Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
            Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH,
            Aircraft.LandingGear.MAIN_GEAR_MASS_SCALER,
            Mission.Design.GROSS_MASS]
        )
        self.declare_partials(Aircraft.LandingGear.NOSE_GEAR_MASS, [
            Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
            Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH,
            Aircraft.LandingGear.NOSE_GEAR_MASS_SCALER,
            Mission.Design.GROSS_MASS]
        )

    def compute(self, inputs, outputs):
        main_gear_length = inputs[Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH]
        main_gear_scaler = inputs[Aircraft.LandingGear.MAIN_GEAR_MASS_SCALER]
        nose_gear_length = inputs[Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH]
        nose_gear_scaler = inputs[Aircraft.LandingGear.NOSE_GEAR_MASS_SCALER]
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM

        total_gear_weight = gross_weight * (
            (30100.0 +
                0.3876 * main_gear_length * main_gear_length +
                0.09579 * nose_gear_length * nose_gear_length
             ) / 1.0e6
        )

        outputs[Aircraft.LandingGear.MAIN_GEAR_MASS] = 0.85 * \
            total_gear_weight * main_gear_scaler / GRAV_ENGLISH_LBM
        outputs[Aircraft.LandingGear.NOSE_GEAR_MASS] = 0.15 * \
            total_gear_weight * nose_gear_scaler / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        main_gear_length = inputs[Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH]
        main_gear_scaler = inputs[Aircraft.LandingGear.MAIN_GEAR_MASS_SCALER]
        nose_gear_length = inputs[Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH]
        nose_gear_scaler = inputs[Aircraft.LandingGear.NOSE_GEAR_MASS_SCALER]
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM

        total_gear_fact = (30100.0 +
                           0.3876 * main_gear_length * main_gear_length +
                           0.09579 * nose_gear_length * nose_gear_length
                           ) / 1.0e6
        total_gear_weight = gross_weight * total_gear_fact
        total_gear_weight_dmain = gross_weight * 7.752e-7 * main_gear_length
        total_gear_weight_dnose = gross_weight * 1.9158e-7 * nose_gear_length

        J[
            Aircraft.LandingGear.MAIN_GEAR_MASS,
            Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH] = (
            0.85 * total_gear_weight_dmain * main_gear_scaler / GRAV_ENGLISH_LBM
        )

        J[
            Aircraft.LandingGear.MAIN_GEAR_MASS,
            Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH] = (
            0.85 * total_gear_weight_dnose * main_gear_scaler / GRAV_ENGLISH_LBM
        )

        J[
            Aircraft.LandingGear.MAIN_GEAR_MASS,
            Aircraft.LandingGear.MAIN_GEAR_MASS_SCALER] = (
            0.85 * total_gear_weight / GRAV_ENGLISH_LBM
        )

        J[Aircraft.LandingGear.MAIN_GEAR_MASS, Mission.Design.GROSS_MASS] = (
            0.85 * total_gear_fact * main_gear_scaler
        )

        J[
            Aircraft.LandingGear.NOSE_GEAR_MASS,
            Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH] = (
            0.15 * total_gear_weight_dmain * nose_gear_scaler / GRAV_ENGLISH_LBM
        )

        J[
            Aircraft.LandingGear.NOSE_GEAR_MASS,
            Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH] = (
            0.15 * total_gear_weight_dnose * nose_gear_scaler / GRAV_ENGLISH_LBM
        )

        J[
            Aircraft.LandingGear.NOSE_GEAR_MASS,
            Aircraft.LandingGear.NOSE_GEAR_MASS_SCALER] = (
            0.15 * total_gear_weight / GRAV_ENGLISH_LBM
        )

        J[Aircraft.LandingGear.NOSE_GEAR_MASS, Mission.Design.GROSS_MASS] = (
            0.15 * total_gear_fact * nose_gear_scaler
        )


class NoseGearLength(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH, val=0.0)
        add_aviary_output(self, Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH, val=0.0)

    def setup_partials(self):
        self.declare_partials(Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH,
                              Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                              val=0.7)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        outputs[Aircraft.LandingGear.NOSE_GEAR_OLEO_LENGTH] = 0.7 * \
            inputs[Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH]


class MainGearLength(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuselage.LENGTH, val=0.0)
        add_aviary_input(self, Aircraft.Fuselage.MAX_WIDTH, val=0.0)
        add_aviary_input(self, Aircraft.Nacelle.AVG_DIAMETER, val=0.0)
        add_aviary_input(self, Aircraft.Engine.WING_LOCATIONS, val=0.0)
        add_aviary_input(self, Aircraft.Wing.DIHEDRAL, val=0.0)
        add_aviary_input(self, Aircraft.Wing.SPAN, val=0.0)

        add_aviary_output(self, Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH, val=0.0)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        options: AviaryValues = self.options['aviary_options']
        # TODO temp using first engine, multi-engine not supported
        num_eng = options.get_val(Aircraft.Engine.NUM_ENGINES)[0]
        num_wing_eng = options.get_val(Aircraft.Engine.NUM_WING_ENGINES)

        y_eng_fore = inputs[Aircraft.Engine.WING_LOCATIONS][0]

        # TODO: high engine-count configuation.
        y_eng_aft = 0

        if num_wing_eng > 0:
            tan_dih = np.tan(inputs[Aircraft.Wing.DIHEDRAL] * DEG2RAD)
            fuse_half_width = inputs[Aircraft.Fuselage.MAX_WIDTH] * 6.0

            d_nacelle = inputs[Aircraft.Nacelle.AVG_DIAMETER]
            # f_nacelle = d_nacelle
            # if num_eng > 4:
            #     f_nacelle = 0.5 * d_nacelle * num_eng ** 0.5

            f_nacelle = distributed_nacelle_diam_factor(d_nacelle, num_eng)

            yee = y_eng_fore
            if num_wing_eng > 2 and y_eng_aft > 0.0:
                yee = y_eng_aft

            if yee < 1.0:
                # This is triggered when the input engine locations are normalized.
                yee *= 6.0 * inputs[Aircraft.Wing.SPAN]

            cmlg = 12.0 * f_nacelle + (0.26 - tan_dih) * (yee - fuse_half_width)

        else:
            cmlg = 0.0

        if cmlg < 12.0:
            cmlg = 0.75 * inputs[Aircraft.Fuselage.LENGTH]

        outputs[Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH] = cmlg

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        options: AviaryValues = self.options['aviary_options']
        # TODO temp using first engine, multi-engine not supported
        num_eng = options.get_val(Aircraft.Engine.NUM_ENGINES)[0]
        num_wing_eng = options.get_val(Aircraft.Engine.NUM_WING_ENGINES)

        y_eng_fore = inputs[Aircraft.Engine.WING_LOCATIONS][0]
        y_eng_aft = 0

        if num_wing_eng > 0:
            tan_dih = np.tan(inputs[Aircraft.Wing.DIHEDRAL] * DEG2RAD)
            dtan_dih = DEG2RAD / np.cos(inputs[Aircraft.Wing.DIHEDRAL] * DEG2RAD) ** 2

            fuse_half_width = inputs[Aircraft.Fuselage.MAX_WIDTH] * 6.0
            dhw_dfuse_wid = 6.0

            d_nacelle = inputs[Aircraft.Nacelle.AVG_DIAMETER]
            # f_nacelle = d_nacelle
            # d_nac = 1.0
            # if num_eng > 4:
            #     f_nacelle = 0.5 * d_nacelle * num_eng ** 0.5
            #     d_nac = 0.5 * num_eng ** 0.5

            f_nacelle = distributed_nacelle_diam_factor(d_nacelle, num_eng)
            d_nac = distributed_nacelle_diam_factor_deriv(num_eng)

            yee = y_eng_fore
            if num_wing_eng > 2 and y_eng_aft > 0.0:
                yee = y_eng_aft

            dyee_dwel = 1.0
            dyee_dspan = 1.0
            if yee < 1.0:
                dyee_dwel = 6.0 * inputs[Aircraft.Wing.SPAN]
                dyee_dspan = 6.0 * yee

                yee *= 6.0 * inputs[Aircraft.Wing.SPAN]

            cmlg = 12.0 * f_nacelle + (0.26 - tan_dih) * (yee - fuse_half_width)
            dcmlg_dnac = 12.0 * d_nac
            dcmlg_dtan = -(yee - fuse_half_width)
            dcmlg_dyee = (0.26 - tan_dih)
            dcmlg_dhw = (tan_dih - 0.26)

        else:
            cmlg = 0.0

        if cmlg < 12.0:
            partials[
                Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                Aircraft.Fuselage.LENGTH] = 0.75

            partials[
                Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                Aircraft.Fuselage.MAX_WIDTH] = 0.0

            partials[
                Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                Aircraft.Nacelle.AVG_DIAMETER] = 0.0

            partials[
                Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                Aircraft.Engine.WING_LOCATIONS] = 0.0

            partials[
                Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                Aircraft.Wing.DIHEDRAL] = 0.0

            partials[Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH, Aircraft.Wing.SPAN] = \
                0.0

        else:
            partials[
                Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH, Aircraft.Fuselage.LENGTH] = \
                0.0

            partials[
                Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                Aircraft.Fuselage.MAX_WIDTH] = dcmlg_dhw * dhw_dfuse_wid

            partials[
                Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                Aircraft.Nacelle.AVG_DIAMETER] = dcmlg_dnac

            partials[
                Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                Aircraft.Engine.WING_LOCATIONS] = dcmlg_dyee * dyee_dwel

            partials[
                Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH,
                Aircraft.Wing.DIHEDRAL] = dcmlg_dtan * dtan_dih

            partials[Aircraft.LandingGear.MAIN_GEAR_OLEO_LENGTH, Aircraft.Wing.SPAN] = \
                dcmlg_dyee * dyee_dspan



from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import shutil
from tkinter import messagebox
import tkinter as tk
from tkinter import messagebox, font as tkFont
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backend_bases import MouseButton
import subprocess
import os


class DraggablePoints:
    def __init__(self, points, ax, canvas):
        self.points = points
        self.ax = ax
        self.canvas = canvas
        self.current_point = None
        self.lines, = ax.plot([p[0] for p in points], [p[1] for p in points],
                              linestyle='-', marker='o', color='b', markersize=8)

        self.cid_click = ax.figure.canvas.mpl_connect(
            'button_press_event', self.on_click)
        self.cid_release = ax.figure.canvas.mpl_connect(
            'button_release_event', self.on_release)
        self.cid_motion = ax.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)

    def on_click(self, event):
        if event.inaxes != self.ax or event.button is not MouseButton.LEFT:
            return
        for i, point in enumerate(self.points):
            if abs(point[0] - event.xdata) < 5.0 and abs(point[1] - event.ydata) < 5.0:
                self.current_point = i
                return

    def on_release(self, event):
        self.current_point = None

    def on_motion(self, event):
        if self.current_point is not None and event.xdata is not None and event.ydata is not None:
            self.points[self.current_point] = (
                self.points[self.current_point][0], event.ydata)
            self.update_points(self.ax, self.points)

    def update_points(self, ax, data):
        color = 'b'
        label = 'Mach Number'
        ax.clear()
        ax.plot([p[0] for p in data], [p[1] for p in data], linestyle='-',
                marker='o', color=color, markersize=8, label=label)

        ax.set_xlim(0, 360)
        ax.set_ylim(0, 1.0)

        ax.figure.canvas.draw()


def create_phase_info(times, altitudes, mach_values,
                      polynomial_order, num_segments, optimize_mach_phase_vars, optimize_altitude_phase_vars, user_choices):
    """
    Creates a dictionary containing the information about different flight phases
    based on input times, altitudes, and Mach values.

    The information includes details such as duration bounds, initial guesses,
    and various options for optimization and control for each phase.

    Parameters
    ----------
    times : list of float
        The times at which phase changes occur, given in minutes.
    altitudes : list of float
        The altitudes corresponding to each phase, given in feet.
    mach_values : list of float
        The Mach numbers corresponding to each phase.

    Returns
    -------
    dict
        A dictionary with all the phase information, including bounds and initial guesses.
    """
    num_phases = len(times) - 1  # Number of phases is one less than the number of points
    phase_info = {}

    times = np.round(np.array(times)).astype(int)
    altitudes = np.round(np.array(altitudes) / 500) * 500
    mach_values = np.round(np.array(mach_values), 2)

    # Utility function to create bounds
    def create_bounds(center):
        lower_bound = max(center / 2, 0.1)  # Ensuring lower bound is not less than 0.1
        upper_bound = center * 1.5
        return (lower_bound, upper_bound)

    # Calculate duration bounds for each phase
    duration_bounds = [create_bounds(times[i+1] - times[i])
                       for i in range(num_phases)]

    # Initialize the cumulative initial bounds
    cumulative_initial_bounds = [(0., 0.)]  # Initial bounds for the first phase

    # Calculate the cumulative initial bounds for subsequent phases
    for i in range(1, num_phases):
        previous_duration_bounds = duration_bounds[i-1]
        previous_initial_bounds = cumulative_initial_bounds[-1]
        new_initial_bound_min = previous_initial_bounds[0] + previous_duration_bounds[0]
        new_initial_bound_max = previous_initial_bounds[1] + previous_duration_bounds[1]
        cumulative_initial_bounds.append((new_initial_bound_min, new_initial_bound_max))

    # Add pre_mission and post_mission phases
    phase_info['pre_mission'] = {
        'include_takeoff': False,
        'optimize_mass': True,
    }

    climb_count = 1
    cruise_count = 1
    descent_count = 1

    for i in range(num_phases):
        initial_altitude = altitudes[i]
        final_altitude = altitudes[i+1]

        # Determine phase type: climb, cruise, or descent
        if final_altitude > initial_altitude:
            phase_type = 'climb'
            phase_count = climb_count
            climb_count += 1
        elif final_altitude == initial_altitude:
            phase_type = 'cruise'
            phase_count = cruise_count
            cruise_count += 1
        else:
            phase_type = 'descent'
            phase_count = descent_count
            descent_count += 1

        phase_name = f'{phase_type}_{phase_count}'

        phase_info[phase_name] = {
            'subsystem_options': {
                'core_aerodynamics': {'method': 'computed'}
            },
            'user_options': {
                'optimize_mach': optimize_mach_phase_vars[i].get(),
                'optimize_altitude': optimize_altitude_phase_vars[i].get(),
                'polynomial_control_order': polynomial_order,
                'use_polynomial_control': True,
                'num_segments': num_segments,
                'order': 3,
                'solve_for_distance': False,
                'initial_mach': (mach_values[i], 'unitless'),
                'final_mach': (mach_values[i+1], 'unitless'),
                'mach_bounds': ((np.min(mach_values[i:i+2]) - 0.02, np.max(mach_values[i:i+2]) + 0.02), 'unitless'),
                'initial_altitude': (altitudes[i], 'ft'),
                'final_altitude': (altitudes[i+1], 'ft'),
                'altitude_bounds': ((max(np.min(altitudes[i:i+2]) - 500., 0.), np.max(altitudes[i:i+2]) + 500.), 'ft'),
                'throttle_enforcement': 'path_constraint' if (i == (num_phases - 1) or i == 0) else 'boundary_constraint',
                'fix_initial': True if i == 0 else False,
                'constrain_final': True if i == (num_phases - 1) else False,
                'fix_duration': False,
                'initial_bounds': (cumulative_initial_bounds[i], 'min'),
                'duration_bounds': (duration_bounds[i], 'min'),
            },
            'initial_guesses': {
                'times': ([times[i], times[i+1]-times[i]], 'min'),
            }
        }

    phase_info['post_mission'] = {
        'include_landing': False,
        'constrain_range': True,
        'target_range': (0., 'nmi'),
    }

    # Apply user choices to each phase
    for phase_name, phase_data in phase_info.items():
        if 'pre_mission' in phase_name or 'post_mission' in phase_name:
            continue
        phase_info[phase_name]['user_options'].update({
            'solve_for_distance': user_choices.get('solve_for_distance', False),
        })

    # Apply global settings if required
    phase_info['post_mission']['constrain_range'] = user_choices.get(
        'constrain_range', True)

    # Calculate the total range
    total_range = estimate_total_range_trapezoidal(times, mach_values)
    print(
        f"Total range is estimated to be {total_range} nautical miles")

    phase_info['post_mission']['target_range'] = (total_range, 'nmi')

    filename = os.path.join(os.getcwd(), 'outputted_phase_info.py')

    # write a python file with the phase information
    with open(filename, 'w') as f:
        f.write(f'phase_info = {phase_info}')

    # Check for 'black' and format the file
    if shutil.which('black'):
        subprocess.run(['black', filename])
    else:
        if shutil.which('autopep8'):
            subprocess.run(['autopep8', '--in-place', '--aggressive', filename])
            print("File formatted using 'autopep8'")
        else:
            print("'black' and 'autopep8' are not installed. Please consider installing one of them for better formatting.")

    print(f"Phase info has been saved and formatted in {filename}")

    return phase_info


def estimate_total_range_trapezoidal(times, mach_numbers):
    speed_of_sound = 343  # Speed of sound in meters per second

    # Convert times to seconds from minutes
    times_sec = np.array(times) * 60

    # Calculate the speeds at each Mach number
    speeds = np.array(mach_numbers) * speed_of_sound

    # Use numpy's trapz function to integrate
    total_range = np.trapz(speeds, times_sec)

    total_range_nautical_miles = total_range / 1000 / 1.852
    return int(round(total_range_nautical_miles))


class Tooltip:
    def __init__(self, ax, tooltip_func):
        self.ax = ax
        self.tooltip_func = tooltip_func
        self.tooltip_widget = None
        self._id = None

    def __call__(self, event):
        if event.inaxes == self.ax:
            x, y = event.xdata, event.ydata
            label = self.tooltip_func(x, y)
            if label:
                if not self.tooltip_widget:
                    self.tooltip_widget = tk.Toplevel()
                    self.tooltip_widget.wm_overrideredirect(True)
                    self.label = tk.Label(self.tooltip_widget, text="", justify='left',
                                          background='white', relief='solid', borderwidth=2,
                                          font=("tahoma", "14", "normal"))
                    self.label.pack(ipadx=1)
                self.label.config(text=label)
                self.tooltip_widget.geometry(
                    f"+{event.guiEvent.x_root+10}+{event.guiEvent.y_root+10}")
                self.tooltip_widget.deiconify()
            else:
                if self.tooltip_widget:
                    self.tooltip_widget.withdraw()
        else:
            if self.tooltip_widget:
                self.tooltip_widget.withdraw()


class IntegratedPlottingApp(tk.Tk):

    def __init__(self):
        super().__init__()
        self.title('Flight Profile Utility')
        self.geometry('1200x900')

        # Set default font for the application
        default_font = tkFont.nametofont("TkDefaultFont")
        default_font.configure(size=18)  # You can adjust the size as needed
        self.option_add("*Font", default_font)

        # You might also want to adjust the font of the Matplotlib plots:
        plt.rc('font', size=12)  # Adjust the size as needed for Matplotlib plots

        # Initialize data
        self.altitude_data = [(0, 0)]
        self.mach_data = [(0, 0.2)]

        # Frame for plots
        plot_frame = tk.Frame(self)
        plot_frame.pack(side='left', fill='both', expand=True)

        # Create Altitude Plot
        self.fig_altitude, self.ax_altitude = plt.subplots()
        self.setup_plot(self.ax_altitude)
        self.canvas_altitude = FigureCanvasTkAgg(self.fig_altitude, master=plot_frame)
        self.canvas_altitude.get_tk_widget().pack(side='top', fill='both', expand=1)

        # Create Mach Plot
        self.fig_mach, self.ax_mach = plt.subplots()
        self.setup_plot(self.ax_mach)
        self.canvas_mach = FigureCanvasTkAgg(self.fig_mach, master=plot_frame)
        self.canvas_mach.get_tk_widget().pack(side='top', fill='both', expand=1)
        self.draggable_points_mach = DraggablePoints(
            self.mach_data, self.ax_mach, self.canvas_mach)

        # Frame for checkboxes
        checkbox_frame = tk.Frame(self)
        checkbox_frame.pack(side='right', fill='y', expand=False)

        # Checkboxes
        self.optimize_mach_var = tk.BooleanVar()
        self.constrain_range_var = tk.BooleanVar()
        tk.Checkbutton(checkbox_frame, text="Constrain Range",
                       variable=self.constrain_range_var).pack(anchor="w")
        self.constrain_range_var.set(True)
        self.solve_for_distance_var = tk.BooleanVar()
        tk.Checkbutton(checkbox_frame, text="Solve for Range",
                       variable=self.solve_for_distance_var).pack(anchor="w")

        # Textbox for Polynomial Control Order
        self.polynomial_order_var = tk.StringVar()
        tk.Label(checkbox_frame, text="Polynomial Control Order").pack(anchor="w")
        self.polynomial_order_entry = tk.Entry(
            checkbox_frame, textvariable=self.polynomial_order_var)
        self.polynomial_order_entry.pack(anchor="w")
        self.polynomial_order_entry.insert(0, "1")

        # Textbox for Number of Segments
        self.num_segments_var = tk.StringVar(value="2")  # Default value set to "2"
        tk.Label(checkbox_frame, text="Number of Segments").pack(anchor="w")
        self.num_segments_entry = tk.Entry(
            checkbox_frame, textvariable=self.num_segments_var)
        self.num_segments_entry.pack(anchor="w")

        self.cid_click_altitude = self.fig_altitude.canvas.mpl_connect(
            'button_press_event', self.on_click)
        self.cid_motion_altitude = self.fig_altitude.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)
        self.cid_release_altitude = self.fig_altitude.canvas.mpl_connect(
            'button_release_event', self.on_release)

        # Done button
        self.done_button = tk.Button(checkbox_frame, text="Done", command=self.on_done)
        self.done_button.pack(side='top', pady=10)

        # Help button
        self.help_button = tk.Button(checkbox_frame, text="Help", command=self.show_help)
        self.help_button.pack(side='top', pady=10)

        # Frame for point entries
        self.point_entry_frame = tk.Frame(checkbox_frame)
        self.point_entry_frame.pack(side='top', expand=False)
        self.altitude_entries = []
        self.mach_entries = []

        # Column labels
        tk.Label(self.point_entry_frame, text="Altitude").grid(row=0, column=1)
        tk.Label(self.point_entry_frame, text="Mach").grid(row=0, column=3)

        # Frame for phase-specific checkboxes
        self.phase_option_frame = tk.Frame(checkbox_frame)
        self.phase_option_frame.pack(side='top', expand=False)

        self.optimize_mach_phase_vars = []
        self.optimize_altitude_phase_vars = []

        # Tooltips for both plots
        self.tooltip_altitude = Tooltip(self.ax_altitude, self.tooltip_func)
        self.tooltip_mach = Tooltip(self.ax_mach, self.tooltip_func)

        self.cid_motion_altitude = self.fig_altitude.canvas.mpl_connect(
            'motion_notify_event', self.tooltip_altitude)
        self.cid_motion_mach = self.fig_mach.canvas.mpl_connect(
            'motion_notify_event', self.tooltip_mach)

        # Bind Enter key to the on_done function
        self.bind("<Return>", self.on_done)

        self.update_point_entries()

        self.current_point = None

    def tooltip_func(self, x, y):
        # Function to generate tooltip text
        for point in self.altitude_data:
            if abs(point[0] - x) < 5.0 and abs(point[1] - y) < 1000.0:
                return f"Time: {int(point[0])} min, Altitude: {int(point[1])} ft"
        for point in self.mach_data:
            if abs(point[0] - x) < 5.0 and abs(point[1] - y) < 0.1:
                return f"Time: {int(point[0])} min, Mach: {round(point[1], 2)}"
        return None

    def update_point_entries(self):
        # Clear existing entries
        for entry in self.altitude_entries + self.mach_entries:
            entry.destroy()
        self.altitude_entries.clear()
        self.mach_entries.clear()

        # Update phase-specific checkboxes
        self.update_phase_options()

        # Create new entries for each point
        for i, (alt, mach) in enumerate(zip(self.altitude_data, self.mach_data)):
            # Altitude Entry
            tk.Label(self.point_entry_frame,
                     text=f"Pt. {i+1}").grid(row=i+1, column=0, sticky="w")
            alt_var = tk.StringVar(value=str(alt[1]))
            alt_entry = tk.Entry(self.point_entry_frame,
                                 textvariable=alt_var, width=10)  # Width set to 10
            alt_entry.grid(row=i+1, column=1)
            alt_entry.bind("<KeyRelease>", lambda e, i=i,
                           var=alt_var: self.update_point_data(i, 'altitude', var.get()))

            # Mach Entry
            tk.Label(self.point_entry_frame,
                     text=f"Pt. {i+1}").grid(row=i+1, column=2, sticky="w")
            mach_var = tk.StringVar(value=str(mach[1]))
            mach_entry = tk.Entry(self.point_entry_frame,
                                  textvariable=mach_var, width=10)  # Width set to 10
            mach_entry.grid(row=i+1, column=3)
            mach_entry.bind("<KeyRelease>", lambda e, i=i,
                            var=mach_var: self.update_point_data(i, 'mach', var.get()))

            self.altitude_entries.append(alt_entry)
            self.mach_entries.append(mach_entry)

    def update_phase_options(self):
        tk.Label(self.phase_option_frame, text=f"Optimize:").grid(
            row=0, column=0, columnspan=3, sticky="s")
        # Create new checkboxes for each phase
        for i in range(len(self.altitude_data) - 1):
            if i >= len(self.optimize_mach_phase_vars):
                mach_var = tk.BooleanVar(value=False)
                alt_var = tk.BooleanVar(value=False)
                self.optimize_mach_phase_vars.append(mach_var)
                self.optimize_altitude_phase_vars.append(alt_var)

            tk.Label(self.phase_option_frame,
                     text=f"Phase {i + 1}").grid(row=i+1, column=0, sticky="w")
            tk.Checkbutton(self.phase_option_frame, text="Altitude",
                           variable=self.optimize_altitude_phase_vars[i]).grid(row=i+1, column=1)
            tk.Checkbutton(self.phase_option_frame, text="Mach",
                           variable=self.optimize_mach_phase_vars[i]).grid(row=i+1, column=2)

    def update_point_data(self, index, point_type, value):
        try:
            value = float(value)
        except ValueError:
            messagebox.showerror("Input Error", "Please enter a valid number.")
            return

        if point_type == 'altitude':
            self.altitude_data[index] = (self.altitude_data[index][0], value)
            self.update_plot(self.ax_altitude, self.altitude_data, 'Altitude', 'g')
        elif point_type == 'mach':
            self.mach_data[index] = (self.mach_data[index][0], value)
            self.update_plot(self.ax_mach, self.mach_data, 'Mach Number', 'b')

    def on_done(self, event=None):  # event=None to handle both button click and keypress
        times, altitudes = zip(*self.altitude_data)
        _, mach_values = zip(*self.mach_data)

        # Retrieve user choices from checkboxes
        user_choices = {
            "constrain_range": self.constrain_range_var.get(),
            "solve_for_distance": self.solve_for_distance_var.get()
        }

        polynomial_order = int(self.polynomial_order_var.get())
        num_segments = int(self.num_segments_var.get())

        # Call create_phase_info and close the window
        # Modify the phase_info creation to include phase-specific options
        create_phase_info(times, altitudes, mach_values, polynomial_order, num_segments,
                          self.optimize_mach_phase_vars, self.optimize_altitude_phase_vars, user_choices)

        self.destroy()

    def show_help(self):
        help_message = (
            "How to Use:\n"
            "- Click on the Altitude plot to add points.\n"
            "- Drag points in the Mach plot to adjust values vertically.\n"
            "- Use the checkboxes on the right to set options for the flight profile:\n"
            "   - Optimize Mach: If checked, the flight profile will optimize for Mach number.\n"
            "   - Optimize Altitude: If checked, the flight profile will optimize for altitude efficiency.\n"
            "   - Constrain Range: If checked, the flight profile will include constraints on the flight range.\n"
            "   - Solve for Range: If checked, the application will calculate the total flight range.\n"
            "- Press 'Done' or the Enter key to process data and close the application.\n\n"
            "Note: Dragging points in the plots will only change their vertical position"
            "while keeping the time (horizontal position) constant."
        )
        messagebox.showinfo("Help", help_message)

    def setup_plot(self, ax):
        if ax == self.ax_altitude:
            title = 'Altitude Profile'
            xlabel = 'Time (minutes)'
            ylabel = 'Altitude (feet)'
            xlim_min = 0
            xlim_max = 360
            ylim_min = 0
            ylim_max = 50000
        elif ax == self.ax_mach:
            title = 'Mach Profile'
            xlabel = 'Time (minutes)'
            ylabel = 'Mach'
            ylim_min = 0
            ylim_max = 1.0

        xlim_min = 0
        xlim_max = 360
        ax.set_title(title)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_xlim(xlim_min, xlim_max)
        ax.set_ylim(ylim_min, ylim_max)

    def on_click(self, event):
        self.update_point_entries()
        if event.inaxes != self.ax_altitude:
            return  # Only interact with the altitude plot

        self.current_point = None
        # Check if the click is near an existing point
        for i, point in enumerate(self.altitude_data):
            if abs(point[0] - event.xdata) < 5.0 and abs(point[1] - event.ydata) < 1000.0:
                self.current_point = i
                return

        # Add a new point only if the click is to the right of the last point
        if event.xdata > max(self.altitude_data, key=lambda x: x[0])[0]:
            x = round(event.xdata)
            y = round(event.ydata / 500) * \
                500 if event.inaxes == self.ax_altitude else round(
                    event.ydata / 0.02) * 0.02

            self.altitude_data.append((x, y))

            # Optionally, add a corresponding point on the Mach plot at the same x
            default_mach_y = 0.72
            self.mach_data.append((x, default_mach_y))

            # Update both plots
            self.update_plot(self.ax_altitude, self.altitude_data, 'Altitude', 'g')
            self.update_plot(self.ax_mach, self.mach_data, 'Mach Number', 'b')

    def on_motion(self, event):
        if self.current_point is not None and event.xdata is not None and event.ydata is not None:
            self.altitude_data[self.current_point] = (
                self.altitude_data[self.current_point][0], event.ydata)
            self.update_plot(self.ax_altitude, self.altitude_data, 'Altitude', 'g')

    def on_release(self, event):
        self.update_point_entries()
        self.current_point = None

    def update_plot(self, ax, data, label, color):
        ax.clear()
        ax.plot([p[0] for p in data], [p[1] for p in data], linestyle='-',
                marker='o', color=color, markersize=8, label=label)

        self.setup_plot(ax)
        ax.figure.canvas.draw()


def _setup_flight_profile_parser(parser):
    """
    Set up the command line options for the Flight Profile plotting tool.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser instance.
    """
    pass


def _exec_flight_profile(options, user_args):
    """
    Run the Flight Profile plotting tool.

    Parameters
    ----------
    options : argparse.Namespace
        Command line options.
    user_args : list of str
        Args to be passed to the user script.
    """
    app = IntegratedPlottingApp()
    app.mainloop()


if __name__ == "__main__":
    app = IntegratedPlottingApp()
    app.mainloop()



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.mass.flops_based.distributed_prop import nacelle_count_factor
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class NacelleMass(om.ExplicitComponent):
    '''
    Calculates total mass of all nacelles in a set for an engine model.
    The methodology is based on the FLOPS weight equations, modified to
    output mass instead of weight.

    Assumptions
    -----------
    all engines have identical nacelles, so Nacelle values are being used for propulsion
    -level totals as well as totals for a set of engines for an engine model
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Nacelle.AVG_DIAMETER, val=np.array([0.0]))

        add_aviary_input(self, Aircraft.Nacelle.AVG_LENGTH, val=np.array([0.0]))

        add_aviary_input(self, Aircraft.Nacelle.MASS_SCALER, val=np.array([1.0]))

        add_aviary_input(self, Aircraft.Engine.SCALED_SLS_THRUST, val=np.array([0.0]))

        add_aviary_output(self, Aircraft.Nacelle.MASS, val=np.array([0.0]))

    def setup_partials(self):
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_eng = aviary_options.get_val(Aircraft.Engine.NUM_ENGINES)
        avg_diam = inputs[Aircraft.Nacelle.AVG_DIAMETER]
        avg_length = inputs[Aircraft.Nacelle.AVG_LENGTH]
        scaler = inputs[Aircraft.Nacelle.MASS_SCALER]

        count_factor = nacelle_count_factor(num_eng)
        # TODO: This should be distributed thrust factor
        thrust = inputs[Aircraft.Engine.SCALED_SLS_THRUST]

        outputs[Aircraft.Nacelle.MASS] = 0.25 * count_factor * \
            avg_diam * avg_length * thrust**0.36 * scaler / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_eng = aviary_options.get_val(Aircraft.Engine.NUM_ENGINES)
        avg_diam = inputs[Aircraft.Nacelle.AVG_DIAMETER]
        avg_length = inputs[Aircraft.Nacelle.AVG_LENGTH]
        scaler = inputs[Aircraft.Nacelle.MASS_SCALER]

        count_factor = nacelle_count_factor(num_eng)
        # TODO: This should be distributed thrust factor
        thrust = inputs[Aircraft.Engine.SCALED_SLS_THRUST]

        thrust_exp = thrust**0.36

        J[Aircraft.Nacelle.MASS, Aircraft.Nacelle.AVG_DIAMETER] = 0.25 * \
            count_factor * avg_length * thrust_exp * scaler / GRAV_ENGLISH_LBM
        J[Aircraft.Nacelle.MASS, Aircraft.Nacelle.AVG_LENGTH] = 0.25 * \
            count_factor * avg_diam * thrust_exp * scaler / GRAV_ENGLISH_LBM
        J[Aircraft.Nacelle.MASS, Aircraft.Nacelle.MASS_SCALER] = 0.25 * \
            count_factor * avg_diam * avg_length * thrust_exp / GRAV_ENGLISH_LBM
        J[Aircraft.Nacelle.MASS, Aircraft.Engine.SCALED_SLS_THRUST] = 0.09 * \
            count_factor * avg_diam * avg_length * thrust**-0.64 * scaler / GRAV_ENGLISH_LBM



import argparse
import os
import sys

import aviary
from aviary.interface.methods_for_level1 import _exec_level1, _setup_level1_parser
from aviary.utils.Fortran_to_Aviary import _exec_F2A, _setup_F2A_parser
from aviary.visualization.dashboard import _dashboard_setup_parser, _dashboard_cmd
from aviary.interface.graphical_input import _exec_flight_profile, _setup_flight_profile_parser
from aviary.interface.download_models import _exec_hangar, _setup_hangar_parser


def _load_and_exec(script_name, user_args):
    """
    Load and exec the given script as __main__.

    Parameters
    ----------
    script_name : str
        The name of the script to load and exec.
    user_args : list of str
        Args to be passed to the user script.
    """

    sys.path.insert(0, os.path.dirname(script_name))

    sys.argv[:] = [script_name] + user_args

    with open(script_name, 'rb') as fp:
        code = compile(fp.read(), script_name, 'exec')

    globals_dict = {
        '__file__': script_name,
        '__name__': '__main__',
        '__package__': None,
        '__cached__': None,
    }

    exec(code, globals_dict)  # nosec: private, internal use only


_command_map = {
    'fortran_to_aviary': (_setup_F2A_parser, _exec_F2A,
                          "Converts legacy Fortran input decks to Aviary csv based decks"),
    'run_mission': (_setup_level1_parser, _exec_level1,
                    "Runs Aviary using a provided input deck"),
    'draw_mission': (_setup_flight_profile_parser, _exec_flight_profile,
                     "Allows users to draw a mission profile for use in Aviary."),
    'dashboard': (_dashboard_setup_parser, _dashboard_cmd,
                  "Run the Dashboard tool"),
    'hangar': (_setup_hangar_parser, _exec_hangar,
               "Allows users that pip installed Aviary to download models from the Aviary hangar"),
}


def aviary_cmd():
    """
    Run an 'aviary' sub-command or list help info for 'aviary' command or sub-commands.
    """
    # pre-parse sys.argv to split between before and after '--'
    if '--' in sys.argv:
        idx = sys.argv.index('--')
        sys_args = sys.argv[:idx]
        user_args = sys.argv[idx + 1:]
        sys.argv[:] = sys_args
    else:
        user_args = []

    parser = argparse.ArgumentParser(description='aviary Command Line Tools',
                                     #  epilog='Use -h after any sub-command for sub-command help, '
                                     #  'for example, "openmdao tree -h" for help on the "tree" '
                                     #  'command. If using a tool on a script that takes its own '
                                     #  'command line arguments, place those arguments after a "--". '
                                     #  'For example: '
                                     #  '"openmdao n2 -o foo.html myscript.py -- -x --myarg=bar"'
                                     )

    # Adding the --version argument
    parser.add_argument('--version', action='store_true', help='show version and exit')

    subs = parser.add_subparsers(title='Tools', metavar='', dest="subparser_name")
    for p, (parser_setup_func, executor, help_str) in sorted(_command_map.items()):
        subp = subs.add_parser(p, help=help_str)
        parser_setup_func(subp)
        subp.set_defaults(executor=executor)

    args = [a for a in sys.argv[1:] if not a.startswith('-')]
    # '--version', '--dependency_versions')]
    cmdargs = [a for a in sys.argv[1:] if a not in ('-h',)]

    if len(args) == 1 and len(user_args) == 0:
        if args[0] != 'draw_mission' and args[0] != 'run_mission':
            parser.parse_args([args[0], '-h'])

    if not set(args).intersection(subs.choices) and len(args) == 1 and os.path.isfile(cmdargs[0]):
        _load_and_exec(args[0], user_args)
    else:
        options, unknown = parser.parse_known_args()

        # Check if --version was passed
        if options.version:
            print(f"Aviary version: {aviary.__version__}")
            return

        if unknown:
            msg = 'unrecognized arguments: ' + ', '.join(unknown)
            try:
                sub = subs.choices[options.subparser_name]
            except KeyError:
                parser.error(msg)
            else:
                print(sub.format_usage(), file=sys.stderr)
                print(msg, file=sys.stderr)
            parser.exit(2)

        if hasattr(options, 'executor'):
            options.executor(options, user_args)
        else:
            os.system('aviary -h')


if __name__ == '__main__':
    aviary_cmd()



import numpy as np
import openmdao.api as om

from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Dynamic


class BasicFlapsCalculations(om.ExplicitComponent):
    def setup(self):

        # inputs

        add_aviary_input(self, Aircraft.Wing.SWEEP, val=25.0)
        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, val=10.13)
        add_aviary_input(self, Aircraft.Wing.FLAP_CHORD_RATIO, val=0.3)
        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, val=0.33)
        add_aviary_input(self, Aircraft.Wing.CENTER_CHORD, val=17.48974)
        add_aviary_input(self, Aircraft.Fuselage.AVG_DIAMETER, val=13.1)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15)
        add_aviary_input(self, Aircraft.Wing.SPAN, val=117.8)
        add_aviary_input(self, Aircraft.Wing.SLAT_CHORD_RATIO, val=0.15)
        self.add_input(
            "slat_defl",
            val=10.0,
            units="deg",
            desc="DELLED: leading edge slat deflection",
        )
        add_aviary_input(self, Aircraft.Wing.OPTIMUM_SLAT_DEFLECTION, val=20.0)
        self.add_input(
            "flap_defl",
            val=10.0,
            units="deg",
            desc="DFLPTO | DFLPLD: takeoff or landing flap deflection",
        )
        add_aviary_input(self, Aircraft.Wing.OPTIMUM_FLAP_DEFLECTION, val=55.0)
        add_aviary_input(self, Aircraft.Wing.ROOT_CHORD, val=16.406626)
        add_aviary_input(self, Aircraft.Fuselage.LENGTH, val=129.4)
        add_aviary_input(self, Aircraft.Wing.LEADING_EDGE_SWEEP,
                         val=0.47639, units="rad")

        # outputs

        self.add_output(
            "VLAM8",
            val=0.74444322,
            units='unitless',
            desc="VLAM8: sensitivity of flap clean wing maximum lift coefficient to wing sweep angle",
        )
        self.add_output(
            "VDEL4",
            val=0.93577864,
            units='unitless',
            desc="VDEL4: sensitivity of flap minimum drag coefficient to flap hinge line sweep",
        )
        self.add_output(
            "VDEL5",
            val=0.90759603,
            units='unitless',
            desc="VDEL5: sensitivity of minimum drag coefficient to fuselage width to span ratio",
        )
        self.add_output(
            "VLAM9",
            val=0.9975,
            units='unitless',
            desc="VLAM9: sensitivity of slat clean wing maximum lift coefficient to slat chord",
        )
        self.add_output(
            "slat_defl_ratio",
            val=0.5,
            units='unitless',
            desc="RDELL: ratio of leading edge slat deflection to optimum deflection angle",
        )
        self.add_output(
            "flap_defl_ratio",
            val=0.5,
            units='unitless',
            desc="RDELF: ratio of trailing edge flap deflection to optimum deflection angle",
        )
        add_aviary_output(self, Aircraft.Wing.SLAT_SPAN_RATIO, 0.89759603)
        self.add_output(
            "chord_to_body_ratio",
            val=0.12679,
            units='unitless',
            desc="CROBL: ratio of root chord to fuselage length",
        )
        self.add_output(
            "body_to_span_ratio",
            val=0.09240397,
            units='unitless',
            desc="FWOB: ratio of body diameter at quarter chord to wing span",
        )
        self.add_output(
            "VLAM12",
            val=0.79207395,
            units='unitless',
            desc="VLAM12: sensitivity of slat clean wing maximum lift coefficient to leading edge sweepback",
        )

    def setup_partials(self):

        # output partials
        self.declare_partials(
            "VLAM8", [Aircraft.Wing.SWEEP], dependent=True, method="cs", step=1e-8
        )
        self.declare_partials(
            "VDEL4",
            [
                Aircraft.Wing.SWEEP,
                Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Wing.FLAP_CHORD_RATIO,
                Aircraft.Wing.TAPER_RATIO,
            ],
            dependent=True,
            method="cs",
            step=1e-8,
        )
        self.declare_partials(
            "VDEL5",
            [
                Aircraft.Wing.SPAN,
                Aircraft.Wing.THICKNESS_TO_CHORD_ROOT,
                Aircraft.Wing.CENTER_CHORD,
                Aircraft.Fuselage.AVG_DIAMETER,
            ],
            dependent=True,
            method="cs",
            step=1e-8,
        )
        self.declare_partials(
            "VLAM9", [Aircraft.Wing.SLAT_CHORD_RATIO], dependent=True, method="cs", step=1e-8
        )
        self.declare_partials(
            "slat_defl_ratio",
            ["slat_defl", Aircraft.Wing.OPTIMUM_SLAT_DEFLECTION],
            dependent=True,
            method="cs",
            step=1e-8,
        )
        self.declare_partials(
            "flap_defl_ratio", ["flap_defl", Aircraft.Wing.OPTIMUM_FLAP_DEFLECTION], method="cs"
        )
        self.declare_partials(
            Aircraft.Wing.SLAT_SPAN_RATIO,
            [
                Aircraft.Wing.SPAN,
                Aircraft.Wing.THICKNESS_TO_CHORD_ROOT,
                Aircraft.Wing.CENTER_CHORD,
                Aircraft.Fuselage.AVG_DIAMETER,
            ],
            dependent=True,
            method="cs",
            step=1e-8,
        )
        self.declare_partials(
            "chord_to_body_ratio",
            [Aircraft.Wing.ROOT_CHORD, Aircraft.Fuselage.LENGTH],
            dependent=True,
            method="cs",
            step=1e-8,
        )
        self.declare_partials(
            "body_to_span_ratio",
            [
                Aircraft.Wing.SPAN,
                Aircraft.Wing.THICKNESS_TO_CHORD_ROOT,
                Aircraft.Wing.CENTER_CHORD,
                Aircraft.Fuselage.AVG_DIAMETER,
            ],
            dependent=True,
            method="cs",
            step=1e-8,
        )
        self.declare_partials(
            "VLAM12",
            [Aircraft.Wing.LEADING_EDGE_SWEEP],
            dependent=True,
            method="cs",
            step=1e-8,
        )

    def compute(self, inputs, outputs):

        sweep_c4 = inputs[Aircraft.Wing.SWEEP]
        AR = inputs[Aircraft.Wing.ASPECT_RATIO]
        flap_chord_ratio = inputs[Aircraft.Wing.FLAP_CHORD_RATIO]
        taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
        center_chord = inputs[Aircraft.Wing.CENTER_CHORD]
        cabin_width = inputs[Aircraft.Fuselage.AVG_DIAMETER]
        tc_ratio_root = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_ROOT]
        wingspan = inputs[Aircraft.Wing.SPAN]
        slat_chord_ratio = inputs[Aircraft.Wing.SLAT_CHORD_RATIO]
        slat_defl = inputs["slat_defl"]
        optimum_slat_defl = inputs[Aircraft.Wing.OPTIMUM_SLAT_DEFLECTION]
        flap_defl = inputs["flap_defl"]
        optimum_flap_defl = inputs[Aircraft.Wing.OPTIMUM_FLAP_DEFLECTION]
        root_chord = inputs[Aircraft.Wing.ROOT_CHORD]
        fus_len = inputs[Aircraft.Fuselage.LENGTH]
        sweep_LE = inputs[Aircraft.Wing.LEADING_EDGE_SWEEP]

        # intermediate values equations
        RLMC4 = sweep_c4 * 0.017453

        TSWPFH = (np.tan(RLMC4)) - (4.0 / AR) * (
            (0.75 - flap_chord_ratio) * (1.0 - taper_ratio) / (1.0 + taper_ratio)
        )
        SWPFHL = np.arctan(TSWPFH)

        DBALE = (
            2.0
            * (
                tc_ratio_root
                * center_chord
                * (cabin_width - (tc_ratio_root * center_chord))
            )
            ** 0.5
        ) + 0.4

        SWPL12 = sweep_LE - 5.0 / 57.296

        # outputs equations
        outputs["VLAM8"] = VLAM8 = (np.cos(RLMC4)) ** 3
        outputs["body_to_span_ratio"] = body_to_span_ratio = DBALE / wingspan
        outputs["VDEL4"] = VDEL4 = np.cos(SWPFHL)
        outputs["VDEL5"] = VDEL5 = 1.0 - body_to_span_ratio
        outputs["VLAM9"] = VLAM9 = 6.65 * slat_chord_ratio
        outputs["slat_defl_ratio"] = slat_defl_ratio = slat_defl / optimum_slat_defl
        outputs["flap_defl_ratio"] = flap_defl / optimum_flap_defl
        outputs[Aircraft.Wing.SLAT_SPAN_RATIO] = slat_span_ratio = 0.99 - DBALE / wingspan
        outputs["chord_to_body_ratio"] = chord_to_body_ratio = root_chord / fus_len
        outputs["VLAM12"] = VLAM12 = (np.cos(SWPL12)) ** 3



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.geometry.gasp_based.non_dimensional_conversion import \
    DimensionalNonDimensionalInterchange
from aviary.variable_info.variables import Aircraft

from aviary.subsystems.geometry.gasp_based.non_dimensional_conversion import DimensionalNonDimensionalInterchange
from aviary.variable_info.options import get_option_defaults


class FoldOnlyTestCase1(unittest.TestCase):
    def setUp(self):
        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')
        options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem("dimensionless_calcs",
                                      DimensionalNonDimensionalInterchange(
                                          aviary_options=options),
                                      promotes_inputs=["aircraft:*"],
                                      promotes_outputs=["aircraft:*"]
                                      )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=180.0, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDED_SPAN, val=118.0, units="ft"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        self.prob.run_model()
        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.FOLDED_SPAN_DIMENSIONLESS], 0.65556, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class FoldOnlyTestCase2(unittest.TestCase):
    def setUp(self):
        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')
        options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                        val=False, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem("dimensionless_calcs",
                                      DimensionalNonDimensionalInterchange(
                                          aviary_options=options),
                                      promotes_inputs=["aircraft:*"],
                                      promotes_outputs=["aircraft:*"]
                                      )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=150.0, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDED_SPAN_DIMENSIONLESS, val=0.5, units="unitless"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        self.prob.run_model()
        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.FOLDED_SPAN], 75.0, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class StrutOnlyTestCase1(unittest.TestCase):
    def setUp(self):
        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')
        options.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem("dimensionless_calcs",
                                      DimensionalNonDimensionalInterchange(
                                          aviary_options=options),
                                      promotes_inputs=["aircraft:*"],
                                      promotes_outputs=["aircraft:*"]
                                      )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=180.0, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Strut.ATTACHMENT_LOCATION, val=118.0, units="ft"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        self.prob.run_model()
        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Strut.ATTACHMENT_LOCATION_DIMENSIONLESS], 0.65556, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class StrutOnlyTestCase2(unittest.TestCase):
    def setUp(self):
        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')
        options.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED,
                        val=False, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem("dimensionless_calcs",
                                      DimensionalNonDimensionalInterchange(
                                          aviary_options=options),
                                      promotes_inputs=["aircraft:*"],
                                      promotes_outputs=["aircraft:*"]
                                      )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=150.0, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Strut.ATTACHMENT_LOCATION_DIMENSIONLESS, val=0.5, units="unitless"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        self.prob.run_model()
        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Strut.ATTACHMENT_LOCATION], 75.0, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class FoldAndStrutTestCase1(unittest.TestCase):
    def setUp(self):
        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')
        options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')
        options.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED,
                        val=False, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem("dimensionless_calcs",
                                      DimensionalNonDimensionalInterchange(
                                          aviary_options=options),
                                      promotes_inputs=["aircraft:*"],
                                      promotes_outputs=["aircraft:*"]
                                      )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=180.0, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDED_SPAN, val=118.0, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Strut.ATTACHMENT_LOCATION_DIMENSIONLESS, val=0.5, units="unitless"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        self.prob.run_model()
        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.FOLDED_SPAN_DIMENSIONLESS], 0.65556, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Strut.ATTACHMENT_LOCATION], 90.0, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class FoldAndStrutTestCase2(unittest.TestCase):
    def setUp(self):
        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')
        options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                        val=False, units='unitless')
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')
        options.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem("dimensionless_calcs",
                                      DimensionalNonDimensionalInterchange(
                                          aviary_options=options),
                                      promotes_inputs=["aircraft:*"],
                                      promotes_outputs=["aircraft:*"]
                                      )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=150.0, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDED_SPAN_DIMENSIONLESS, val=0.8, units="unitless"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Strut.ATTACHMENT_LOCATION, val=90.0, units="ft"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        self.prob.run_model()
        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.FOLDED_SPAN], 120.0, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Strut.ATTACHMENT_LOCATION_DIMENSIONLESS], 0.6, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


class FoldAndStrutTestCase3(unittest.TestCase):
    def setUp(self):
        options = get_option_defaults()
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')
        options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')
        options.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem("dimensionless_calcs",
                                      DimensionalNonDimensionalInterchange(
                                          aviary_options=options),
                                      promotes_inputs=["aircraft:*"],
                                      promotes_outputs=["aircraft:*"]
                                      )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=180.0, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDED_SPAN, val=118.0, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Strut.ATTACHMENT_LOCATION, val=108.0, units="ft"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        self.prob.run_model()
        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.FOLDED_SPAN_DIMENSIONLESS], 0.65556, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Strut.ATTACHMENT_LOCATION_DIMENSIONLESS], .6, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


if __name__ == "__main__":
    unittest.main()
